import abc
from typing import Literal, overload

import polars as pl

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.riskdataset_settings import (
    DatasetStatus,
    RiskDatasetMetadata,
    RiskDatasetProperties,
    RiskDatasetSettings,
    RiskDatasetSettingsMenu,
    RiskDatasetUpdateResult,
)
from bayesline.api._src.registry import (
    AsyncRegistryBasedApi,
    RawSettings,
    RegistryBasedApi,
)
from bayesline.api._src.tasks import AsyncTask, Task


class DatasetError(Exception):
    """Exception raised for dataset-related errors."""

    ...


class RiskDatasetApi(abc.ABC):
    """API for managing risk datasets.

    This abstract base class defines the interface for risk dataset operations,
    including updating datasets and retrieving their properties.
    """

    @property
    @abc.abstractmethod
    def settings(self) -> RiskDatasetSettings:
        """Get the settings for this risk dataset.

        Returns
        -------
        RiskDatasetSettings
            The settings for this risk dataset.
        """
        ...

    @abc.abstractmethod
    def describe(self) -> RiskDatasetProperties:
        """Get the properties of this risk dataset.

        Returns
        -------
        RiskDatasetProperties
            The properties of this risk dataset.
        """
        ...

    @abc.abstractmethod
    def update(self, force: bool = False) -> RiskDatasetUpdateResult:
        """Check underlying datasets for updates and update this dataset to latest versions.

        Parameters
        ----------
        force : bool, default=False
            If true, the update will be forced even if the dataset is up to date.

        Raises
        ------
        DatasetError
            If an error occurs.

        Returns
        -------
        RiskDatasetUpdateResult
            The result of the update operation.
        """

    @abc.abstractmethod
    def update_as_task(self) -> Task[RiskDatasetUpdateResult]: ...  # noqa: D102


class RiskDatasetLoaderApi(
    RegistryBasedApi[RiskDatasetSettings, RiskDatasetSettingsMenu, RiskDatasetApi],
):
    """API for loading and managing risk datasets.

    This class provides functionality for loading, creating, and managing risk datasets
    through a registry-based approach.
    """

    @abc.abstractmethod
    def get_default_dataset_name(self) -> str:
        """Get the default dataset name.

        Returns
        -------
        str
            The default dataset name that will be populated into settings if no
            dataset name is provided.
        """

    @abc.abstractmethod
    def get_dataset_names(
        self,
        *,
        mode: Literal["System", "User"] | None = None,
    ) -> dict[str, DatasetStatus]:
        """Get the names of all available datasets.

        Parameters
        ----------
        mode : Literal["System", "User"] | None, default=None
            System: only system wide datasets (available to all users and provided
            by the system).
            User: only user specific datasets (available to the current user).
            None: all datasets (system wide and user specific).

        Returns
        -------
        dict[str, DatasetStatus]
            The names of all available datasets and their status.
            "ready" means the dataset is ready to use.
            "data_stale" means the dataset is ready to use but newer data is available.
            "needs_update" means local data needs to be computed/updated.
            "not_ready" means the reference dataset is not in a usable state.
            "corrupt" means the underlying settings are corrupted and need to be repaired.
        """

    @overload
    def list_riskdatasets(
        self, name: str | None = None, *, df: Literal[False] = ...
    ) -> list[RiskDatasetMetadata]: ...

    @overload
    def list_riskdatasets(
        self, name: str | None = None, *, df: Literal[True]
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    def list_riskdatasets(
        self, name: str | None = None, *, df: bool = False
    ) -> list[RiskDatasetMetadata] | pl.DataFrame:
        """List all available risk datasets.

        Parameters
        ----------
        name : str | None, default=None
            The name of the dataset to filter by.
            If not given, all datasets are returned.
        df : bool, default=False
            If True, return a polars DataFrame instead of a list of metadata objects.

        Returns
        -------
        list[RiskDatasetMetadata] | pl.DataFrame
            The metadata of all available datasets, either as a list of objects
            or as a DataFrame depending on the `df` parameter.
        """

    def create_or_replace_dataset(
        self, name: str, settings: RiskDatasetSettings
    ) -> RiskDatasetApi:
        """Create a new dataset or replace an existing one.

        Parameters
        ----------
        name : str
            The name of the dataset to create or replace.
        settings : RiskDatasetSettings
            The settings for the dataset.

        Returns
        -------
        RiskDatasetApi
            The API for the created or replaced dataset.
        """
        self.delete_dataset_if_exists(name)
        return self.create_dataset(name, settings)

    @abc.abstractmethod
    def create_dataset(
        self, name: str, settings: RiskDatasetSettings
    ) -> RiskDatasetApi:
        """Create a new dataset with the given name and settings.

        Parameters
        ----------
        name : str
            The name of the dataset to create.
        settings : RiskDatasetSettings
            The settings for the dataset to create.

        Raises
        ------
        DatasetError
            If a dataset with the given name already exists
            or if the settings are otherwise invalid.

        Returns
        -------
        RiskDatasetApi
            The API of the newly created dataset.
        """

    @abc.abstractmethod
    def create_dataset_as_task(  # noqa: D102
        self, name: str, settings: RiskDatasetSettings
    ) -> Task[RiskDatasetApi]: ...

    @abc.abstractmethod
    def delete_dataset(self, name: str) -> RawSettings:
        """Delete the given dataset.

        Parameters
        ----------
        name : str
            The name of the dataset to delete.

        Raises
        ------
        KeyError
            If the dataset does not exist.
        DatasetError
            If the dataset could not be deleted.

        Returns
        -------
        RawSettings
            The raw settings of the deleted dataset.
        """

    def delete_dataset_if_exists(self, name: str) -> RawSettings | None:
        """Delete the given dataset if it exists.

        Parameters
        ----------
        name : str
            The name of the dataset to delete.

        Raises
        ------
        DatasetError
            If the dataset could not be deleted.

        Returns
        -------
        RawSettings | None
            The raw settings of the deleted dataset if dataset existed.
        """
        if name in self.settings.names(mode="All"):
            return self.delete_dataset(name)
        return None


@docstrings_from_sync
class AsyncRiskDatasetApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> RiskDatasetSettings: ...  # noqa: D102

    @abc.abstractmethod
    async def describe(self) -> RiskDatasetProperties: ...  # noqa: D102

    @abc.abstractmethod
    async def update(  # noqa: D102
        self, force: bool = False
    ) -> RiskDatasetUpdateResult: ...

    @abc.abstractmethod
    async def update_as_task(  # noqa: D102
        self, force: bool = False
    ) -> AsyncTask[RiskDatasetUpdateResult]: ...


@docstrings_from_sync
class AsyncRiskDatasetLoaderApi(
    AsyncRegistryBasedApi[
        RiskDatasetSettings, RiskDatasetSettingsMenu, AsyncRiskDatasetApi
    ],
):

    @abc.abstractmethod
    async def get_default_dataset_name(self) -> str: ...  # noqa: D102

    @abc.abstractmethod
    async def get_dataset_names(  # noqa: D102
        self,
        *,
        mode: Literal["System", "User"] | None = None,
    ) -> dict[str, DatasetStatus]: ...

    @overload
    async def list_riskdatasets(
        self, name: str | None = None, *, df: Literal[False] = ...
    ) -> list[RiskDatasetMetadata]: ...

    @overload
    async def list_riskdatasets(
        self, name: str | None = None, *, df: Literal[True]
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def list_riskdatasets(  # noqa: D102
        self, name: str | None = None, *, df: bool = False
    ) -> list[RiskDatasetMetadata] | pl.DataFrame: ...

    async def create_or_replace_dataset(  # noqa: D102
        self, name: str, settings: RiskDatasetSettings
    ) -> AsyncRiskDatasetApi:
        await self.delete_dataset_if_exists(name)
        return await self.create_dataset(name, settings)

    @abc.abstractmethod
    async def create_dataset(  # noqa: D102
        self, name: str, settings: RiskDatasetSettings
    ) -> AsyncRiskDatasetApi: ...

    @abc.abstractmethod
    async def create_dataset_as_task(  # noqa: D102
        self, name: str, settings: RiskDatasetSettings
    ) -> AsyncTask[AsyncRiskDatasetApi]: ...

    @abc.abstractmethod
    async def delete_dataset(self, name: str) -> RawSettings: ...  # noqa: D102

    async def delete_dataset_if_exists(  # noqa: D102
        self, name: str
    ) -> RawSettings | None:
        if name in await self.settings.names(mode="All"):
            return await self.delete_dataset(name)
        return None



import abc
import datetime as dt
from typing import Any, Literal

import polars as pl
from pydantic import Field

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.settings import ReportSettings
from bayesline.api._src.equity.riskmodels_settings import FactorRiskModelSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike


class FactorStatsReportSettings(ReportSettings):
    """Settings for a factor stats report."""

    report_type: Literal["factor_stats"] = Field(
        default="factor_stats", description="Report type discriminator"
    )
    factor_model_settings: FactorRiskModelSettings = Field()


class FactorStatsReportAccessor(TypedReportAccessorApi):
    """Specific accessor for a exposure report."""

    @abc.abstractmethod
    def get_dates(self) -> list[dt.date]:
        """
        Get the dates covered in the factor stats report.

        Returns
        -------
        list[dt.date]
            The dates covered in the factor stats report.
        """

    @abc.abstractmethod
    def get_factors(self) -> dict[str, list[str]]:
        """
        Get the factor groups and their factor covered in this factor stats report.

        Returns
        -------
        dict[str, list[str]]
            Dict where the keys are the factor groups, and the values are the factors
            in that group.
        """

    @abc.abstractmethod
    def get_factor_returns(self) -> pl.DataFrame:
        """
        Get the time-series of factor returns for the factors.

        Returns
        -------
        pl.DataFrame
            A dataframe with a `date` column and columns for each factor with their
            respective factor returns, in factor_group^factor format.
        """

    @abc.abstractmethod
    def get_t_stats(self) -> pl.DataFrame:
        """
        Get the time-series of t-statistics for the factors.

        Returns
        -------
        pl.DataFrame
            A dataframe with a `date` column and columns for each factor with their
            respective t-statistics, in factor_group^factor format.
        """

    @abc.abstractmethod
    def get_p_values(self) -> pl.DataFrame:
        """
        Get the time-series of p-values for the factors.

        Returns
        -------
        pl.DataFrame
            A dataframe with a `date` column and columns for each factor with their
            respective p-values, in factor_group^factor format.
        """

    @abc.abstractmethod
    def get_vif(self) -> pl.DataFrame:
        r"""
        Get the time-series of variance inflation factors (VIF) for the factors.

        The VIF measures multicollinearity — how much the variance of a
        regression coefficient is inflated due to correlation with other
        factors. For factor *j*, VIF is defined as:

        .. math::

            \\text{VIF}_j = \\frac{1}{1 - R_j^2}

        where :math:`R_j^2` is the R-squared from regressing factor *j* on all
        other factors. Equivalently,
        :math:`\\text{VIF}_j = (X^\\top W X)_{jj} \\cdot (X^\\top W X)^{-1}_{jj}`.

        A VIF of 1 means no multicollinearity; values above 5--10 suggest that
        the factor's coefficient estimate is unreliable due to near-linear
        dependence on other factors.

        When the model includes linear constraints (e.g., a sum-to-zero
        constraint on industry factor returns), the unconstrained Gram matrix
        :math:`X^\\top W X` is typically singular or near-singular along the
        constrained directions — for instance, the market intercept column is
        a linear combination of the industry dummy columns. In this case VIF
        is computed in the null space of the constraint matrix, measuring
        multicollinearity only among the degrees of freedom that the model is
        actually estimating. Factors whose collinearity is resolved by the
        constraint will show moderate VIF rather than artificially extreme
        values.

        Returns
        -------
        pl.DataFrame
            A dataframe with a `date` column and columns for each factor with
            their respective VIF values, in factor_group^factor format.
        """


@docstrings_from_sync
class AsyncFactorStatsReportAccessor(AsyncTypedReportAccessorApi):

    @abc.abstractmethod
    async def get_dates(self) -> list[dt.date]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_factors(self) -> dict[str, list[str]]: ...  # noqa: D102

    @abc.abstractmethod
    async def get_factor_returns(self) -> pl.DataFrame: ...  # noqa: D102

    @abc.abstractmethod
    async def get_t_stats(self) -> pl.DataFrame: ...  # noqa: D102

    @abc.abstractmethod
    async def get_p_values(self) -> pl.DataFrame: ...  # noqa: D102

    @abc.abstractmethod
    async def get_vif(self) -> pl.DataFrame: ...  # noqa: D102


class FactorStatsReportApi(
    ReportApi[
        [DateLike, DateLike], FactorStatsReportAccessor, FactorStatsReportSettings
    ]
):
    """API for a exposure report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> FactorStatsReportAccessor | None:
        """
        Get the cached factor stats report accessor API if it is available.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        FactorStatsReportAccessor | None
            The cached factor stats report accessor API, or None if the report is not cached.
        """

    @abc.abstractmethod
    def calculate(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> FactorStatsReportAccessor:
        """
        Calculate the factor stats report.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        FactorStatsReportAccessor
            The exposure report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[FactorStatsReportAccessor]:
        """
        Calculate the factor stats report as a task.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[FactorStatsReportAccessor]
            The task for the factor stats report accessor.
        """


@docstrings_from_sync
class AsyncFactorStatsReportApi(
    AsyncReportApi[
        [DateLike | None, DateLike | None],
        AsyncFactorStatsReportAccessor,
        FactorStatsReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncFactorStatsReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncFactorStatsReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncFactorStatsReportAccessor]: ...


class FactorStatsReportAccessorImpl(FactorStatsReportAccessor):  # noqa: D101

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor

    def get_dates(self) -> list[dt.date]:  # noqa: D102
        df = self.accessor.get_data(
            path=[],
            expand=("date",),
            value_cols=(),
        )
        return df["date"].unique().to_list()

    def get_factors(self) -> dict[str, list[str]]:  # noqa: D102
        df = self.accessor.get_data(
            [],
            expand=("factor_group", "factor"),
            value_cols=(),
        )
        result: dict[str, list[str]] = {}
        for (factor_group,), group_df in df.group_by("factor_group"):
            result[factor_group] = (
                group_df["factor"].unique(maintain_order=True).to_list()
            )
        return result

    def get_factor_returns(self) -> pl.DataFrame:  # noqa: D102
        df = self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("factor_return",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^factor_return"))

    def get_t_stats(self) -> pl.DataFrame:  # noqa: D102
        df = self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("t_statistic",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^t_statistic"))

    def get_p_values(self) -> pl.DataFrame:  # noqa: D102
        df = self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("p_value",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^p_value"))

    def get_vif(self) -> pl.DataFrame:  # noqa: D102
        df = self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("vif",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^vif"))


class AsyncFactorStatsReportAccessorImpl(AsyncFactorStatsReportAccessor):  # noqa: D101

    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor

    async def get_dates(self) -> list[dt.date]:  # noqa: D102
        df = await self.accessor.get_data(
            path=[],
            expand=("date",),
            value_cols=(),
        )
        return df["date"].unique().to_list()

    async def get_factors(self) -> dict[str, list[str]]:  # noqa: D102
        df = await self.accessor.get_data(
            [],
            expand=("factor_group", "factor"),
            value_cols=(),
        )
        result: dict[str, list[str]] = {}
        for (factor_group,), group_df in df.group_by("factor_group"):
            result[factor_group] = (
                group_df["factor"].unique(maintain_order=True).to_list()
            )
        return result

    async def get_factor_returns(self) -> pl.DataFrame:  # noqa: D102
        df = await self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("factor_return",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^factor_return"))

    async def get_t_stats(self) -> pl.DataFrame:  # noqa: D102
        df = await self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("t_statistic",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^t_statistic"))

    async def get_p_values(self) -> pl.DataFrame:  # noqa: D102
        df = await self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("p_value",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^p_value"))

    async def get_vif(self) -> pl.DataFrame:  # noqa: D102
        df = await self.accessor.get_data(
            [],
            expand=("date",),
            value_cols=("vif",),
            pivot_cols=("factor_group", "factor"),
        )
        return df.rename(lambda x: x.removesuffix("^vif"))



from __future__ import annotations

from collections import Counter
from itertools import pairwise
from typing import Annotated, Any

import polars as pl
from pydantic import Field, ValidationInfo, field_validator, model_validator

from bayesline.api._src.equity.portfolio_settings import PortfolioOrganizerSettings
from bayesline.api._src.registry import Settings, SettingsMenu, SettingsTypeMetaData


class PortfolioHierarchySettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """Specifies the set of available portfolios that can be used to create hierarchies."""

    sources: dict[str, list[str]] = Field(
        default_factory=dict,
        description=(
            "Mapping of sources to the available portfolio IDs for that source."
        ),
    )

    schemas: dict[str, list[str]] = Field(
        default_factory=dict,
        description=(
            "Set of available portfolio organizer schemas and their "
            "associated portfolio IDs."
        ),
    )

    def describe(self) -> str:
        """Describe the available portfolio schemas.

        Returns
        -------
        str
            A description of the available portfolio schemas.
        """
        sources = {k: sorted(v) for k, v in self.sources.items()}
        schemas = {k: sorted(v) for k, v in self.schemas.items()}
        return f"Available Portfolio Sources: {sources}\nAvailable Portfolio Schemas: {schemas}"


class PortfolioHierarchySettings(Settings[PortfolioHierarchySettingsMenu]):
    """Specifies portfolio hierarchies with arbitrary groupings (e.g. manager, etc.)."""

    portfolio_schema: Annotated[
        str | int | PortfolioOrganizerSettings,
        Field(
            description=(
                "The portfolio organizer settings to use as an underlying schema of "
                "portfolios. The 'Default' schema is used by default."
            ),
        ),
        SettingsTypeMetaData[str | int | PortfolioOrganizerSettings](
            references=PortfolioOrganizerSettings
        ),
    ]

    groupings: dict[str, list[str]] = Field(default_factory=dict)
    portfolio_ids: list[str]
    benchmark_ids: list[str | None]

    @property
    def menu_type(self) -> type[PortfolioHierarchySettingsMenu]:  # noqa: D102
        return PortfolioHierarchySettingsMenu

    @classmethod
    def from_source(
        cls: type[PortfolioHierarchySettings],
        source: str,
        portfolio_ids: list[str],
        benchmark_ids: list[str | None] | None = None,
        groupings: dict[str, list[str]] | None = None,
        dataset: str | None = None,
    ) -> PortfolioHierarchySettings:
        """Create portfolio hierarchy settings from a source.

        Parameters
        ----------
        source : str
            The name of an upload from the portfolio uploader.
        portfolio_ids : list[str]
            The list of portfolio IDs.
        benchmark_ids : list[str | None] | None, default=None
            The list of benchmark IDs, defaults to None for each portfolio.
        groupings : dict[str, list[str]] | None, default=None
            The groupings dictionary, defaults to empty dict.
        dataset : str | None, default=None
            The risk dataset to use for the security master and asset returns. If None,
            the default dataset is used.

        Returns
        -------
        PortfolioHierarchySettings
            The created portfolio hierarchy settings.
        """
        return cls(
            portfolio_schema=PortfolioOrganizerSettings(
                enabled_portfolios=source, dataset=dataset
            ),
            portfolio_ids=portfolio_ids,
            benchmark_ids=benchmark_ids or [None] * len(portfolio_ids),
            groupings=groupings or {},
        )

    @classmethod
    def from_polars(
        cls: type[PortfolioHierarchySettings],
        df: pl.DataFrame,
        portfolio_schema: str | int | PortfolioOrganizerSettings = "Default",
        portfolio_source: str | None = None,
        dataset: str | None = None,
    ) -> PortfolioHierarchySettings:
        """Create a portfolio hierarchy from a dataframe.

        Must contain a column `portfolio_id` and optionally `benchmark_id`.
        Every other column is interpreted as a grouping. 0 groupings are allowed.
        Index is ignored.

        Parameters
        ----------
        df : pl.DataFrame
            The dataframe to create the hierarchy from.
        portfolio_schema : str | int | PortfolioOrganizerSettings, default="Default"
            The underlying portfolio schema to use.
        portfolio_source : str | None, default=None
            The source to use for the portfolio schema. If not provided then
            the portfolio schema is used. If provided then it will override the
            portfolio schema.
        dataset : str | None, default=None
            The dataset to use for the portfolio schema. If not provided then
            the default dataset is used. Will be ignored if `source` is not
            provided.

        Returns
        -------
        PortfolioHierarchySettings
            The created portfolio hierarchy settings.

        Raises
        ------
        ValueError
            If the portfolio_id column is not found in the dataframe.
        """
        if "portfolio_id" not in df.columns:
            raise ValueError("portfolio_id column not found in the dataframe.")

        groupings = {
            col: df.get_column(col).to_list()
            for col in df.columns
            if col not in {"portfolio_id", "benchmark_id"}
        }
        portfolio_ids = df.get_column("portfolio_id").to_list()

        if "benchmark_id" not in df.columns:
            benchmark_ids = [None] * len(portfolio_ids)
        else:
            benchmark_ids = df.get_column("benchmark_id").to_list()

        if portfolio_source is not None:
            portfolio_schema = PortfolioOrganizerSettings(
                enabled_portfolios=portfolio_source, dataset=dataset
            )

        return cls(
            portfolio_schema=portfolio_schema,
            groupings=groupings,
            portfolio_ids=portfolio_ids,
            benchmark_ids=benchmark_ids,
        )

    def to_polars(self) -> pl.DataFrame:
        """Convert the hierarchy to a polars dataframe.

        The last two columns are the portfolio and benchmark IDs.
        Every column before that is a grouping. 0 groupings are possible.

        Returns
        -------
        pl.DataFrame
            The dataframe representation of the hierarchy, sorted by portfolio ID.
        """
        return pl.DataFrame(self.groupings).with_columns(
            pl.Series("portfolio_id", self.portfolio_ids),
            pl.Series("benchmark_id", self.benchmark_ids),
        )

    @field_validator("portfolio_ids")
    @classmethod
    def _validate_portfolio_ids(
        cls: type[PortfolioHierarchySettings], v: list[str]
    ) -> list[str]:
        if not v:
            raise ValueError("Portfolio IDs must be non-empty.")

        duplicates = [item for item, count in Counter(v).items() if count > 1]
        if duplicates:
            raise ValueError(
                "Portfolio IDs must be unique. "
                f"Found duplicates: {', '.join(duplicates)}"
            )

        if any(p is None or p.strip() == "" for p in v):
            raise ValueError("Portfolio IDs must be non-empty strings.")

        return v

    @field_validator("portfolio_ids")
    @classmethod
    def _validate_benchmark_ids(
        cls: type[PortfolioHierarchySettings], v: list[str]
    ) -> list[str]:
        if not v:
            raise ValueError("Benchmark IDs must be non-empty.")
        return v

    @field_validator("groupings")
    @classmethod
    def _validate_groupings(
        cls: type[PortfolioHierarchySettings], v: dict[str, list[str]]
    ) -> dict[str, list[str]]:
        if not all(len(v1) == len(v2) for v1, v2 in pairwise(v.values())):
            raise ValueError(f"Groupings must have the same length. {v}")

        groups_with_nulls = []
        for group, values in v.items():
            if any(e is None for e in values):
                groups_with_nulls.append(group)
        if groups_with_nulls:
            raise ValueError(
                f"Groupings must not contain null values. Found in: {groups_with_nulls}"
            )

        return v

    @model_validator(mode="before")
    @classmethod
    def _fill_benchmark_ids(cls: type[PortfolioHierarchySettings], v: Any) -> Any:
        if isinstance(v, dict):
            if "benchmark_ids" not in v or v["benchmark_ids"] is None:
                v["benchmark_ids"] = [None] * len(v["portfolio_ids"])
        return v

    @model_validator(mode="after")
    def _validate_dimensions(self) -> PortfolioHierarchySettings:
        if len(self.portfolio_ids) != len(self.benchmark_ids):
            raise ValueError(
                "Portfolio IDs and benchmark IDs must have the same length."
            )

        n = len(self.portfolio_ids)
        mismatches = [g for g in self.groupings.values() if len(g) != n]
        if len(mismatches) > 0:
            raise ValueError(
                "Portfolio IDs, benchmark IDs and groups must have the same length. "
                f"Found mismatches: {mismatches}"
            )
        return self

    @field_validator("portfolio_schema", mode="after")
    @classmethod
    def _validate_portfolio_schema(
        cls,
        portfolio_schema: str | int | PortfolioOrganizerSettings,
        info: ValidationInfo,
    ) -> str | int | PortfolioOrganizerSettings:
        if isinstance(menu := info.context, PortfolioHierarchySettingsMenu):
            if (
                isinstance(portfolio_schema, str)
                and portfolio_schema not in menu.schemas
            ):
                raise ValueError(
                    f"Unknown schema: {portfolio_schema}, available: {menu.schemas.keys()}"
                )
        return portfolio_schema

    @model_validator(mode="after")
    def _validate_ids_with_rest(
        self, info: ValidationInfo
    ) -> PortfolioHierarchySettings:
        if isinstance(menu := info.context, PortfolioHierarchySettingsMenu):
            if isinstance(self.portfolio_schema, str):
                available_portfolio_ids = set(menu.schemas[self.portfolio_schema])
                available_portfolio_ids_with_rest = {
                    n.split(":")[0] + ":{REST}"
                    for n in available_portfolio_ids
                    if ":" in n
                } | available_portfolio_ids
                portfolio_ids = set(self.portfolio_ids)
                benchmark_ids = {b for b in self.benchmark_ids if b is not None}

                if not portfolio_ids.issubset(available_portfolio_ids_with_rest):
                    raise ValueError(
                        f"There are unknown portfolio ids. "
                        f"Unknown: {portfolio_ids - available_portfolio_ids}"
                    )

                if not benchmark_ids.issubset(available_portfolio_ids):
                    raise ValueError(
                        f"There are unknown benchmark ids. "
                        f"Unknown: {benchmark_ids - available_portfolio_ids}"
                    )
        return self

    def describe(self, menu: PortfolioHierarchySettingsMenu) -> str:
        """Describe the portfolio hierarchy settings.

        Parameters
        ----------
        menu : PortfolioHierarchySettingsMenu
            The menu to get context information from.

        Returns
        -------
        str
            The description of the portfolio hierarchy settings.
        """
        schema = self.portfolio_schema
        if isinstance(schema, str) and schema not in menu.schemas:
            return f"Unknown schema: {schema}"

        df = self.to_polars()
        if isinstance(schema, str):
            df = df.with_columns(
                pl.col("portfolio_id").map_elements(
                    lambda x: f"[{x}]" if x not in menu.schemas[schema] else x,
                    pl.String,
                ),
                pl.col("benchmark_id").map_elements(
                    lambda x: (
                        f"[{x}]"
                        if x is not None and x not in menu.schemas[schema]
                        else x
                    ),
                    pl.String,
                ),
            )
            return f"""Missing portfolios are enclosed in brackets.

{df.__repr__()}"""
        else:
            return df.__repr__()



import abc
from typing import Any, Literal

import polars as pl
from pydantic import Field

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.settings import ReportSettings
from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike


class UniverseCountReportSettings(ReportSettings):
    """Settings for a universe count report."""

    report_type: Literal["universe_count"] = Field(
        default="universe_count", description="Report type discriminator"
    )
    universe_settings: UniverseSettings = Field()
    date_aggregation: Literal["last", "mean", "sum"] = Field(
        default="last", description="The date aggregation to use"
    )


class UniverseCountReportAccessor(TypedReportAccessorApi):
    """Specific accessor for a universe count report."""

    @abc.abstractmethod
    def get_counts(  # noqa: D102
        self,
        *,
        dates: bool = True,
        categorical_hierarchy_levels: dict[str, int] | None = None,
    ) -> pl.DataFrame:
        """
        Get the counts for the universe.

        Parameters
        ----------
        dates : bool, default=True
            If True, groups by dates.
        categorical_hierarchy_levels : dict[str, int] | None, default=None
            The level of categorical aggregation to group by.
            The key is the hierarchy name, the value is the level.
            If None, no categorical aggregation is done. A level of -1 means
            to use all levels.


        Returns
        -------
        pl.DataFrame
            A dataframe with `[date, count]` columns.
        """


@docstrings_from_sync
class AsyncUniverseCountReportAccessor(AsyncTypedReportAccessorApi):

    @abc.abstractmethod
    async def get_counts(  # noqa: D102
        self,
        *,
        dates: bool = True,
        categorical_hierarchy_levels: dict[str, int] | None = None,
    ) -> pl.DataFrame: ...


class UniverseCountReportApi(
    ReportApi[
        [DateLike | None, DateLike | None],
        UniverseCountReportAccessor,
        UniverseCountReportSettings,
    ]
):
    """API for a universe count report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> UniverseCountReportAccessor | None:
        """
        Get the cached universe count report accessor API if it is available.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        UniverseCountReportAccessor | None
            The cached universe count report accessor API, or None if the report is
            not cached.
        """

    @abc.abstractmethod
    def calculate(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> UniverseCountReportAccessor:
        """
        Calculate the universe count report.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        UniverseCountReportAccessor
            The universe count report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[UniverseCountReportAccessor]:
        """
        Calculate the universe count report as a task.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[UniverseCountReportAccessor]
            The task for the universe count report.
        """


@docstrings_from_sync
class AsyncUniverseCountReportApi(
    AsyncReportApi[
        [DateLike | None, DateLike | None],
        AsyncUniverseCountReportAccessor,
        UniverseCountReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncUniverseCountReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncUniverseCountReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncUniverseCountReportAccessor]: ...


class UniverseCountReportAccessorImpl(UniverseCountReportAccessor):  # noqa: D101

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor

    def get_counts(  # noqa: D102
        self,
        *,
        dates: bool = True,
        categorical_hierarchy_levels: dict[str, int] | None = None,
    ) -> pl.DataFrame:
        expand = ["date"] if dates else []
        if categorical_hierarchy_levels:
            for name, max_level in categorical_hierarchy_levels.items():
                for name_level in self.accessor.axes["asset"]:
                    if not name_level.startswith(f"{name}_level_"):
                        continue
                    _, level = name_level.split("_level_")
                    if max_level == -1 or int(level) <= max_level:
                        expand.append(name_level)
        df = self.accessor.get_data([], expand=tuple(expand), value_cols=("count",))
        return df.filter(pl.col("count") != 0)


class AsyncUniverseCountReportAccessorImpl(  # noqa: D101
    AsyncUniverseCountReportAccessor
):
    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor

    async def get_counts(  # noqa: D102
        self,
        *,
        dates: bool = True,
        categorical_hierarchy_levels: dict[str, int] | None = None,
    ) -> pl.DataFrame:
        expand = ["date"] if dates else []
        if categorical_hierarchy_levels:
            for name, max_level in categorical_hierarchy_levels.items():
                for name_level in self.accessor.axes["asset"]:
                    if not name_level.startswith(f"{name}_level_"):
                        continue
                    _, level = name_level.split("_level_")
                    if max_level == -1 or int(level) <= max_level:
                        expand.append(name_level)
        df = await self.accessor.get_data(
            [], expand=tuple(expand), value_cols=("count",)
        )
        return df.filter(pl.col("count") != 0)



import abc
import datetime
from typing import Any

from pydantic import BaseModel

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.types import DateLike


class IncidentSummaryItem(BaseModel):
    """A single incident summary item.

    Parameters
    ----------
    datetime : datetime.datetime
        The datetime when the incident occurred.
    incident_id : str
        The unique identifier for the incident.
    source : str
        The source system that reported the incident.
    """

    datetime: datetime.datetime
    incident_id: str
    source: str


class IncidentSummary(BaseModel):
    """A summary of incidents within a date and index range.

    Parameters
    ----------
    items : list[IncidentSummaryItem]
        List of incident summary items.
    start_date : datetime.datetime
        The start date of the range.
    end_date : datetime.datetime
        The end date of the range.
    n_start : int
        The starting index number.
    n_end : int
        The ending index number.
    n_more : int
        The number of additional incidents beyond the requested range.
    """

    items: list[IncidentSummaryItem]

    start_date: datetime.datetime
    end_date: datetime.datetime
    n_start: int
    n_end: int
    n_more: int


class IncidentsServiceApi(abc.ABC):
    """Gives access to system incidents.

    Includes failed requests, their logs and contextual information.
    """

    @abc.abstractmethod
    def submit_incident(
        self, incident_id: str, source: str, body: dict[str, Any]
    ) -> IncidentSummaryItem:
        r"""Submit an incident with the given ID and source.

        Parameters
        ----------
        incident_id : str
            The ID of the incident.
            Cannot contain any of the following characters: /\\:*?"<>|-
        source : str
            The source of the incident.
            Cannot contain any of the following characters: /\\:*?"<>|-
        body : dict[str, Any]
            The body of the incident.

        Returns
        -------
        IncidentSummaryItem
            The submitted incident item.
        """
        ...

    @abc.abstractmethod
    def get_incident_summary(
        self,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        start_idx: int = 0,
        end_idx: int = 999999999,
    ) -> IncidentSummary:
        """Obtain incident summaries for the given date and index range.

        Parameters
        ----------
        start_date : DateLike | None, default=None
            The start date of the range. If None, the last 24 hours are used.
        end_date : DateLike | None, default=None
            The end date of the range. If None, `now` is used.
        start_idx : int, default=0
            The start index of the range, `0` being first.
        end_idx : int, default=999999999
            The end index of the range, `-1` being last.

        Returns
        -------
        IncidentSummary
            The incident summary.
        """
        ...

    @abc.abstractmethod
    def get_incident(self, incident_id: str) -> dict[str, dict[str, Any]]:
        r"""Obtain the incident with the given ID.

        Parameters
        ----------
        incident_id : str
            The ID of the incident.
            Cannot contain any of the following characters: /\\:*?"<>|-

        Returns
        -------
        dict[str, dict[str, Any]]
            The incident details.
        """
        ...


@docstrings_from_sync
class AsyncIncidentsServiceApi(abc.ABC):

    @abc.abstractmethod
    async def submit_incident(  # noqa: D102
        self, incident_id: str, source: str, body: dict[str, Any]
    ) -> IncidentSummaryItem: ...

    @abc.abstractmethod
    async def get_incident_summary(  # noqa: D102
        self,
        start_date: DateLike | None = None,
        end_date: DateLike | None = None,
        start_idx: int = 0,
        end_idx: int = 999999999,
    ) -> IncidentSummary: ...

    @abc.abstractmethod
    async def get_incident(  # noqa: D102
        self, incident_id: str
    ) -> dict[str, dict[str, Any]]: ...



from collections.abc import Sequence

from bayesline.api._src.equity.calendar_settings import CalendarSettings
from bayesline.api._src.equity.exposure_settings import ExposureSettings
from bayesline.api._src.equity.modelconstruction_settings import (
    ModelConstructionSettings,
)
from bayesline.api._src.equity.portfolio_settings import (
    PortfolioOrganizerSettings,
    PortfolioSettings,
)
from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
)
from bayesline.api._src.equity.report_settings import ReportSettings
from bayesline.api._src.equity.riskdataset_settings import RiskDatasetSettings
from bayesline.api._src.equity.riskmodels_settings import FactorRiskModelSettings
from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.equity.upload_settings import UploadSettings
from bayesline.api._src.registry import Settings

SETTINGS_TYPES: Sequence[type[Settings]] = [
    UniverseSettings,
    PortfolioOrganizerSettings,
    RiskDatasetSettings,
    FactorRiskModelSettings,
    ModelConstructionSettings,
    PortfolioSettings,
    PortfolioHierarchySettings,
    ReportSettings,
    CalendarSettings,
    ExposureSettings,
    UploadSettings,
]



import abc
from typing import Any, Literal

from pydantic import Field

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.calendar_settings import CalendarSettings
from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
)
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.scenarios import PortfolioHoldingsScenario
from bayesline.api._src.equity.report.settings import (
    FactorCovarianceSettings,
    IdioVolSettings,
    ReportSettings,
)
from bayesline.api._src.equity.riskmodels_settings import FactorRiskModelSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike


class XSRReportSettings(ReportSettings):
    """Settings for an XSR (Exposure-Volatility-Correlation-Contribution) report."""

    report_type: Literal["xsr"] = Field(
        default="xsr", description="Report type discriminator"
    )
    portfolio_hierarchy_settings: PortfolioHierarchySettings = Field()
    factor_model_settings: FactorRiskModelSettings = Field()
    factor_cov_settings: FactorCovarianceSettings = Field(
        default_factory=FactorCovarianceSettings,
        description="Settings for the factor covariance matrix forecasts.",
    )
    idio_vol_settings: IdioVolSettings = Field(
        default_factory=IdioVolSettings,
        description="Settings for the idiosyncratic volatility matrix forecasts.",
    )
    calendar: CalendarSettings = Field(default_factory=CalendarSettings)
    currency: str | None = Field(None)
    rescale_bench: bool = Field(True)
    normalize_holdings: bool = Field(True)
    analytics_space: Literal["absolute", "active", "benchmark"] = Field("absolute")
    backfill_holdings: bool = Field(False)

    annualize: bool = Field(default=True)


class XSRReportAccessor(TypedReportAccessorApi):
    """Specific accessor for an XSR report."""

    scenario_types = [PortfolioHoldingsScenario]


@docstrings_from_sync
class AsyncXSRReportAccessor(AsyncTypedReportAccessorApi):
    scenario_types = [PortfolioHoldingsScenario]


class XSRReportApi(
    ReportApi[
        [DateLike | None, DateLike | None],
        XSRReportAccessor,
        XSRReportSettings,
    ]
):
    """API for an XSR report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> XSRReportAccessor | None:
        """
        Get the cached XSR report accessor API if it is available.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        XSRReportAccessor | None
            The XSR report accessor if cached, otherwise None.
        """

    @abc.abstractmethod
    def calculate(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> XSRReportAccessor:
        """
        Calculate the XSR report.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        XSRReportAccessor
            The XSR report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[XSRReportAccessor]:
        """
        Calculate the XSR report as a task.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[XSRReportAccessor]
            The task for the XSR report accessor.
        """


@docstrings_from_sync
class AsyncXSRReportApi(
    AsyncReportApi[
        [DateLike | None, DateLike | None],
        AsyncXSRReportAccessor,
        XSRReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncXSRReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncXSRReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncXSRReportAccessor]: ...


class XSRReportAccessorImpl(XSRReportAccessor):  # noqa: D101

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor


class AsyncXSRReportAccessorImpl(AsyncXSRReportAccessor):  # noqa: D101

    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor



"""Equity API."""

from bayesline.api._src.equity import settings as settings_tools
from bayesline.api._src.equity._status_types import DatasetStatus
from bayesline.api._src.equity.api import AsyncBayeslineEquityApi, BayeslineEquityApi
from bayesline.api._src.equity.calendar import (
    AsyncCalendarApi,
    AsyncCalendarLoaderApi,
    CalendarApi,
    CalendarLoaderApi,
)
from bayesline.api._src.equity.calendar_settings import (
    CalendarSettings,
    CalendarSettingsMenu,
)
from bayesline.api._src.equity.exposure import (
    AsyncExposureApi,
    AsyncExposureLoaderApi,
    ExposureApi,
    ExposureLoaderApi,
)
from bayesline.api._src.equity.exposure_settings import (
    CategoricalExposureGroupSettings,
    ContinuousExposureGroupSettings,
    ExposureGroupSettings,
    ExposureSettings,
    ExposureSettingsMenu,
    HierarchyGroups,
    HierarchyLevel,
    InteractionExposureGroupSettings,
    TSBetaExposureGroupSettings,
)
from bayesline.api._src.equity.ids import (
    AssetIdApi,
    AsyncAssetIdApi,
)
from bayesline.api._src.equity.impl.factormodels_async import (
    AsyncFactorModelEngineImpl,
    AsyncFactorModelImpl,
    AsyncFactorModelLoaderImpl,
)
from bayesline.api._src.equity.impl.factormodels_sync import (
    FactorModelEngineImpl,
    FactorModelImpl,
    FactorModelLoaderImpl,
)
from bayesline.api._src.equity.modelconstruction import (
    AsyncFactorModelConstructionApi,
    AsyncFactorModelConstructionLoaderApi,
    FactorModelConstructionApi,
    FactorModelConstructionLoaderApi,
)
from bayesline.api._src.equity.modelconstruction_settings import (
    ModelConstructionSettings,
    ModelConstructionSettingsMenu,
)
from bayesline.api._src.equity.portfolio import (
    AsyncPortfolioApi,
    AsyncPortfolioLoaderApi,
    PortfolioApi,
    PortfolioLoaderApi,
)
from bayesline.api._src.equity.portfolio_settings import (
    PortfolioOrganizerSettings,
    PortfolioOrganizerSettingsMenu,
    PortfolioSettings,
    PortfolioSettingsMenu,
)
from bayesline.api._src.equity.portfoliohierarchy import (
    AsyncPortfolioHierarchyApi,
    AsyncPortfolioHierarchyLoaderApi,
    PortfolioHierarchyApi,
    PortfolioHierarchyLoaderApi,
)
from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
    PortfolioHierarchySettingsMenu,
)
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    AsyncReportApi,
    AsyncReportLoaderApi,
    AsyncReportPersister,
    IllegalPathError,
    ReportAccessorApi,
    ReportApi,
    ReportLoaderApi,
    ReportPersister,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi as AsyncTypedReportAccessorApiV2,
)
from bayesline.api._src.equity.report.accessor import (
    TypedReportAccessorApi as TypedReportAccessorApiV2,
)
from bayesline.api._src.equity.report.accessors import (
    ACCESSOR_IMPLS,
    ASYNC_ACCESSOR_IMPLS,
)
from bayesline.api._src.equity.report.api import AsyncReportApi as AsyncReportApiV2
from bayesline.api._src.equity.report.api import ReportApi as ReportApiV2
from bayesline.api._src.equity.report.api_loader import (
    AnyReportSettings,
)
from bayesline.api._src.equity.report.api_loader import (
    AsyncReportLoaderApi as AsyncReportLoaderApiV2,
)
from bayesline.api._src.equity.report.api_loader import (
    ReportLoaderApi as ReportLoaderApiV2,
)
from bayesline.api._src.equity.report.covariance import (
    AsyncFactorCovarianceReportAccessor as AsyncFactorCovarianceReportAccessorV2,
)
from bayesline.api._src.equity.report.covariance import (
    AsyncFactorCovarianceReportAccessorImpl as AsyncFactorCovarianceReportAccessorImplV2,
)
from bayesline.api._src.equity.report.covariance import (
    AsyncFactorCovarianceReportApi as AsyncFactorCovarianceReportApiV2,
)
from bayesline.api._src.equity.report.covariance import (
    FactorCovarianceReportAccessorImpl as FactorCovarianceReportAccessorImplV2,
)
from bayesline.api._src.equity.report.covariance import (
    FactorCovarianceReportApi as FactorCovarianceReportApiV2,
)
from bayesline.api._src.equity.report.covariance import (
    FactorCovarianceReportSettings as FactorCovarianceReportSettingsV2,
)
from bayesline.api._src.equity.report.exposure import (
    AsyncExposureReportAccessor as AsyncExposureReportAccessorV2,
)
from bayesline.api._src.equity.report.exposure import (
    AsyncExposureReportAccessorImpl as AsyncExposureReportAccessorImplV2,
)
from bayesline.api._src.equity.report.exposure import (
    AsyncExposureReportApi as AsyncExposureReportApiV2,
)
from bayesline.api._src.equity.report.exposure import (
    ExposureReportAccessorImpl as ExposureReportAccessorImplV2,
)
from bayesline.api._src.equity.report.exposure import (
    ExposureReportApi as ExposureReportApiV2,
)
from bayesline.api._src.equity.report.exposure import (
    ExposureReportSettings as ExposureReportSettingsV2,
)
from bayesline.api._src.equity.report.factor_stats import (
    AsyncFactorStatsReportAccessor as AsyncFactorStatsReportAccessorV2,
)
from bayesline.api._src.equity.report.factor_stats import (
    AsyncFactorStatsReportAccessorImpl as AsyncFactorStatsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.factor_stats import (
    AsyncFactorStatsReportApi as AsyncFactorStatsReportApiV2,
)
from bayesline.api._src.equity.report.factor_stats import (
    FactorStatsReportAccessorImpl as FactorStatsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.factor_stats import (
    FactorStatsReportApi as FactorStatsReportApiV2,
)
from bayesline.api._src.equity.report.factor_stats import (
    FactorStatsReportSettings as FactorStatsReportSettingsV2,
)
from bayesline.api._src.equity.report.idio import (
    AsyncIdioReportAccessor as AsyncIdioReportAccessorV2,
)
from bayesline.api._src.equity.report.idio import (
    AsyncIdioReportAccessorImpl as AsyncIdioReportAccessorImplV2,
)
from bayesline.api._src.equity.report.idio import (
    AsyncIdioReportApi as AsyncIdioReportApiV2,
)
from bayesline.api._src.equity.report.idio import (
    IdioReportAccessorImpl as IdioReportAccessorImplV2,
)
from bayesline.api._src.equity.report.idio import IdioReportApi as IdioReportApiV2
from bayesline.api._src.equity.report.idio import (
    IdioReportSettings as IdioReportSettingsV2,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    AsyncMcapExposureReportAccessor as AsyncMcapExposureReportAccessorV2,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    AsyncMcapExposureReportAccessorImpl as AsyncMcapExposureReportAccessorImplV2,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    AsyncMcapExposureReportApi as AsyncMcapExposureReportApiV2,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    McapExposureReportAccessorImpl as McapExposureReportAccessorImplV2,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    McapExposureReportApi as McapExposureReportApiV2,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    McapExposureReportSettings as McapExposureReportSettingsV2,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    AsyncPortfolioHoldingsReportAccessor as AsyncPortfolioHoldingsReportAccessorV2,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    AsyncPortfolioHoldingsReportAccessorImpl as AsyncPortfolioHoldingsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    AsyncPortfolioHoldingsReportApi as AsyncPortfolioHoldingsReportApiV2,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    PortfolioHoldingsReportAccessorImpl as PortfolioHoldingsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    PortfolioHoldingsReportApi as PortfolioHoldingsReportApiV2,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    PortfolioHoldingsReportSettings as PortfolioHoldingsReportSettingsV2,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    AsyncPortfolioReturnsReportAccessor as AsyncPortfolioReturnsReportAccessorV2,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    AsyncPortfolioReturnsReportAccessorImpl as AsyncPortfolioReturnsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    AsyncPortfolioReturnsReportApi as AsyncPortfolioReturnsReportApiV2,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    PortfolioReturnsReportAccessorImpl as PortfolioReturnsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    PortfolioReturnsReportApi as PortfolioReturnsReportApiV2,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    PortfolioReturnsReportSettings as PortfolioReturnsReportSettingsV2,
)
from bayesline.api._src.equity.report.return_attribution import (
    AsyncReturnAttributionReportAccessor as AsyncReturnAttributionReportAccessorV2,
)
from bayesline.api._src.equity.report.return_attribution import (
    AsyncReturnAttributionReportAccessorImpl as AsyncReturnAttributionReportAccessorImplV2,
)
from bayesline.api._src.equity.report.return_attribution import (
    AsyncReturnAttributionReportApi as AsyncReturnAttributionReportApiV2,
)
from bayesline.api._src.equity.report.return_attribution import (
    ReturnAttributionReportAccessorImpl as ReturnAttributionReportAccessorImplV2,
)
from bayesline.api._src.equity.report.return_attribution import (
    ReturnAttributionReportApi as ReturnAttributionReportApiV2,
)
from bayesline.api._src.equity.report.return_attribution import (
    ReturnAttributionReportSettings as ReturnAttributionReportSettingsV2,
)
from bayesline.api._src.equity.report.scenarios import (
    AnyScenarioSettings,
    PortfolioHoldingsScenario,
    PortfolioHoldingsScenarioOverride,
    ScenarioSettings,
)
from bayesline.api._src.equity.report.settings import (
    FactorCovarianceSettings,
    IdioVolSettings,
)
from bayesline.api._src.equity.report.settings import ReportSettings as ReportSettingsV2
from bayesline.api._src.equity.report.universe import (
    AsyncUniverseCountReportAccessor as AsyncUniverseCountReportAccessorV2,
)
from bayesline.api._src.equity.report.universe import (
    AsyncUniverseCountReportAccessorImpl as AsyncUniverseCountReportAccessorImplV2,
)
from bayesline.api._src.equity.report.universe import (
    AsyncUniverseCountReportApi as AsyncUniverseCountReportApiV2,
)
from bayesline.api._src.equity.report.universe import (
    UniverseCountReportAccessorImpl as UniverseCountReportAccessorImplV2,
)
from bayesline.api._src.equity.report.universe import (
    UniverseCountReportApi as UniverseCountReportApiV2,
)
from bayesline.api._src.equity.report.universe import (
    UniverseCountReportSettings as UniverseCountReportSettingsV2,
)
from bayesline.api._src.equity.report.weights import (
    AsyncWeightsReportAccessor as AsyncWeightsReportAccessorV2,
)
from bayesline.api._src.equity.report.weights import (
    AsyncWeightsReportAccessorImpl as AsyncWeightsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.weights import (
    AsyncWeightsReportApi as AsyncWeightsReportApiV2,
)
from bayesline.api._src.equity.report.weights import (
    WeightsReportAccessorImpl as WeightsReportAccessorImplV2,
)
from bayesline.api._src.equity.report.weights import (
    WeightsReportApi as WeightsReportApiV2,
)
from bayesline.api._src.equity.report.weights import (
    WeightsReportSettings as WeightsReportSettingsV2,
)
from bayesline.api._src.equity.report.xsr import (
    AsyncXSRReportAccessor as AsyncXSRReportAccessorV2,
)
from bayesline.api._src.equity.report.xsr import (
    AsyncXSRReportAccessorImpl as AsyncXSRReportAccessorImplV2,
)
from bayesline.api._src.equity.report.xsr import (
    AsyncXSRReportApi as AsyncXSRReportApiV2,
)
from bayesline.api._src.equity.report.xsr import (
    XSRReportAccessorImpl as XSRReportAccessorImplV2,
)
from bayesline.api._src.equity.report.xsr import XSRReportApi as XSRReportApiV2
from bayesline.api._src.equity.report.xsr import (
    XSRReportSettings as XSRReportSettingsV2,
)
from bayesline.api._src.equity.report_settings import (
    AssetForecastLossReportSettings,
    AssetHoldingsMeasureSettings,
    AssetHoldingsReportSettings,
    AssetHoldingsReportSettingsMenu,
    AssetStressTestMeasureSettings,
    AssetStressTestReportSettings,
    AssetStressTestReportSettingsMenu,
    AVaRMeasureSettings,
    BrinsonAttributionMeasureSettings,
    BrinsonAttributionReportSettings,
    BrinsonAttributionReportSettingsMenu,
    ConcreteReportSettings,
    ConcreteReportSettingsMenu,
    CumsumMeasureSettings,
    DrawdownMeasureSettings,
    ExposureMeasureSettings,
    ExposureReportSettings,
    ExposureReportSettingsMenu,
    FactorAttributionDrilldownMeasureSettings,
    FactorAttributionDrilldownReportSettings,
    FactorAttributionDrilldownReportSettingsMenu,
    FactorAttributionMeasureSettings,
    FactorAttributionReportSettings,
    FactorAttributionReportSettingsMenu,
    FactorCovarianceMeasureSettings,
    FactorCovarianceReportSettings,
    FactorCovarianceReportSettingsMenu,
    FactorForecastLossReportSettings,
    FactorIdioMeasureSettings,
    FactorIdioReportSettings,
    FactorIdioReportSettingsMenu,
    FactorMovingAverageMeasureSettings,
    FactorTSReportSettings,
    FactorVolForecastMeasureSettings,
    ForecastBacktestMeasureSettings,
    ForecastBacktestReportSettings,
    ForecastBacktestReportSettingsMenu,
    HoldingsMeasureSettings,
    HoldingsReportSettings,
    HoldingsReportSettingsMenu,
    HVaRMeasureSettings,
    IdiosyncraticReturnMeasureSettings,
    IdiosyncraticReturnReportSettings,
    IdiosyncraticReturnReportSettingsMenu,
    IdiosyncraticVolatilityMeasureSettings,
    IdiosyncraticVolatilityReportSettings,
    IdiosyncraticVolatilityReportSettingsMenu,
    MeasureSettings,
    MovingAverageMeasureSettings,
    MovingAverageRSquaredMeasureSettings,
    PassThroughFactor2DMeasureSettings,
    PortfolioStressTestMeasureSettings,
    PortfolioStressTestReportSettings,
    PortfolioStressTestReportSettingsMenu,
    ReportAccessorSettings,
    ReportSettings,
    ReportSettingsMenu,
    RiskDecompositionMeasureSettings,
    RiskDecompositionReportSettings,
    RiskDecompositionReportSettingsMenu,
    RiskModelFitFactorReportSettings,
    RiskModelFitFactorReportSettingsMenu,
    RiskModelFitReportSettings,
    RiskModelPortfolioFitReportSettings,
    StyleCorrelationReportSettings,
    StyleCorrelationReportSettingsMenu,
    StyleIndustryExposureReportSettings,
    StyleIndustryExposureReportSettingsMenu,
    TimeSeriesBetaMeasureSettings,
    TimeSeriesVolatilityMeasureSettings,
    TimeSeriesXSigmaRhoMeasureSettings,
    TSBetaReportSettings,
    TSXSRReportSettings,
    TSXSRReportSettingsMenu,
    VaRReportSettings,
    VaRReportSettingsMenu,
    VolForecastMeasureSettings,
    XSigmaRhoMeasureSettings,
    XSRReportSettings,
    XSRReportSettingsMenu,
)
from bayesline.api._src.equity.riskdataset import (
    AsyncRiskDatasetApi,
    AsyncRiskDatasetLoaderApi,
    DatasetError,
    RiskDatasetApi,
    RiskDatasetLoaderApi,
)
from bayesline.api._src.equity.riskdataset_settings import (
    RiskDatasetExposureSettings,
    RiskDatasetHuberRegressionExposureSettings,
    RiskDatasetMetadata,
    RiskDatasetProperties,
    RiskDatasetReferencedExposureSettings,
    RiskDatasetSettings,
    RiskDatasetSettingsMenu,
    RiskDatasetUnitExposureSettings,
    RiskDatasetUpdateResult,
    RiskDatasetUploadedExposureSettings,
)
from bayesline.api._src.equity.riskmodels import (
    AsyncFactorModelApi,
    AsyncFactorModelEngineApi,
    AsyncFactorModelLoaderApi,
    AsyncFactorModelMetadataApi,
    FactorModelApi,
    FactorModelEngineApi,
    FactorModelLoaderApi,
    FactorModelMetadataApi,
)
from bayesline.api._src.equity.riskmodels_settings import (
    FactorRiskModelMetadata,
    FactorRiskModelSettings,
    FactorRiskModelSettingsMenu,
    GetModelMode,
)
from bayesline.api._src.equity.settings import hierarchy_df_to_wide, hierarchy_to_df
from bayesline.api._src.equity.universe import (
    AsyncUniverseApi,
    AsyncUniverseLoaderApi,
    UniverseApi,
    UniverseLoaderApi,
)
from bayesline.api._src.equity.universe_settings import (
    CategoricalFilterSettings,
    Hierarchy,
    MCapFilterSettings,
    UniverseSettings,
    UniverseSettingsMenu,
)
from bayesline.api._src.equity.upload import (
    AsyncDataTypeUploaderApi,
    AsyncUploaderApi,
    AsyncUploadersApi,
    AsyncUploadParserApi,
    DataTypeUploaderApi,
    MultiParserResult,
    UploadCommitResult,
    UploadDataTypeFreshness,
    UploaderApi,
    UploadError,
    UploadersApi,
    UploadParserApi,
    UploadParserResult,
    UploadStagingResult,
)
from bayesline.api._src.equity.upload_settings import (
    UploadSettings,
    UploadSettingsMenu,
)

__all__ = [
    "BayeslineEquityApi",
    "ExposureApi",
    "AsyncExposureApi",
    "UniverseLoaderApi",
    "ExposureLoaderApi",
    "UniverseSettings",
    "UniverseSettingsMenu",
    "CategoricalFilterSettings",
    "MCapFilterSettings",
    "Hierarchy",
    "UniverseApi",
    "AsyncUniverseApi",
    "HierarchyLevel",
    "HierarchyGroups",
    "MeasureSettings",
    "ExposureSettings",
    "ExposureSettingsMenu",
    "ExposureGroupSettings",
    "CategoricalExposureGroupSettings",
    "ContinuousExposureGroupSettings",
    "TSBetaExposureGroupSettings",
    "InteractionExposureGroupSettings",
    "TimeSeriesVolatilityMeasureSettings",
    "TimeSeriesXSigmaRhoMeasureSettings",
    "CumsumMeasureSettings",
    "DrawdownMeasureSettings",
    "MovingAverageMeasureSettings",
    "FactorMovingAverageMeasureSettings",
    "RiskModelFitReportSettings",
    "RiskModelPortfolioFitReportSettings",
    "VolForecastMeasureSettings",
    "FactorVolForecastMeasureSettings",
    "FactorIdioMeasureSettings",
    "MovingAverageRSquaredMeasureSettings",
    "settings_tools",
    "ModelConstructionSettings",
    "ModelConstructionSettingsMenu",
    "ReportSettings",
    "ReportSettingsMenu",
    "ConcreteReportSettings",
    "ConcreteReportSettingsMenu",
    "TSXSRReportSettings",
    "TSXSRReportSettingsMenu",
    "RiskDecompositionMeasureSettings",
    "RiskDecompositionReportSettings",
    "RiskDecompositionReportSettingsMenu",
    "XSRReportSettings",
    "XSRReportSettingsMenu",
    "VaRReportSettings",
    "VaRReportSettingsMenu",
    "ExposureMeasureSettings",
    "FactorTSReportSettings",
    "HVaRMeasureSettings",
    "AVaRMeasureSettings",
    "FactorForecastLossReportSettings",
    "AssetForecastLossReportSettings",
    "ReportApi",
    "AsyncReportApi",
    "ReportLoaderApi",
    "FactorModelConstructionLoaderApi",
    "FactorModelEngineApi",
    "AsyncFactorModelEngineApi",
    "FactorModelApi",
    "AsyncFactorModelApi",
    "FactorRiskModelSettings",
    "FactorRiskModelSettingsMenu",
    "FactorModelLoaderApi",
    "FactorModelConstructionApi",
    "AsyncFactorModelConstructionApi",
    "FactorIdioReportSettings",
    "FactorCovarianceSettings",
    "IdioVolSettings",
    "RiskModelFitFactorReportSettings",
    "RiskModelFitFactorReportSettingsMenu",
    "PassThroughFactor2DMeasureSettings",
    "StyleCorrelationReportSettings",
    "StyleIndustryExposureReportSettings",
    "FactorIdioReportSettingsMenu",
    "StyleCorrelationReportSettingsMenu",
    "StyleIndustryExposureReportSettingsMenu",
    "AsyncBayeslineEquityApi",
    "AsyncExposureLoaderApi",
    "AsyncUniverseLoaderApi",
    "AsyncFactorModelLoaderApi",
    "AsyncFactorModelConstructionLoaderApi",
    "AsyncReportLoaderApi",
    "XSigmaRhoMeasureSettings",
    "ForecastBacktestReportSettings",
    "ForecastBacktestReportSettingsMenu",
    "ForecastBacktestMeasureSettings",
    "BrinsonAttributionMeasureSettings",
    "BrinsonAttributionReportSettings",
    "BrinsonAttributionReportSettingsMenu",
    "FactorAttributionMeasureSettings",
    "FactorAttributionReportSettings",
    "FactorAttributionReportSettingsMenu",
    "FactorAttributionDrilldownMeasureSettings",
    "FactorAttributionDrilldownReportSettings",
    "FactorAttributionDrilldownReportSettingsMenu",
    "PortfolioLoaderApi",
    "AsyncPortfolioLoaderApi",
    "PortfolioStressTestMeasureSettings",
    "PortfolioStressTestReportSettings",
    "PortfolioStressTestReportSettingsMenu",
    "AssetStressTestMeasureSettings",
    "AssetStressTestReportSettings",
    "AssetStressTestReportSettingsMenu",
    "AssetHoldingsMeasureSettings",
    "AssetHoldingsReportSettings",
    "AssetHoldingsReportSettingsMenu",
    "PortfolioHierarchySettings",
    "PortfolioHierarchySettingsMenu",
    "PortfolioHierarchyLoaderApi",
    "AsyncPortfolioHierarchyLoaderApi",
    "AsyncPortfolioHierarchyApi",
    "PortfolioHierarchyApi",
    "HoldingsMeasureSettings",
    "HoldingsReportSettings",
    "HoldingsReportSettingsMenu",
    "AssetIdApi",
    "AsyncAssetIdApi",
    "ExposureMeasureSettings",
    "ExposureReportSettings",
    "ExposureReportSettingsMenu",
    "PortfolioSettings",
    "PortfolioSettingsMenu",
    "AsyncPortfolioApi",
    "PortfolioApi",
    "UploadParserApi",
    "UploadParserResult",
    "AsyncUploadParserApi",
    "PortfolioOrganizerSettings",
    "PortfolioOrganizerSettingsMenu",
    "IdiosyncraticReturnMeasureSettings",
    "IdiosyncraticReturnReportSettings",
    "IdiosyncraticReturnReportSettingsMenu",
    "IdiosyncraticVolatilityMeasureSettings",
    "IdiosyncraticVolatilityReportSettings",
    "IdiosyncraticVolatilityReportSettingsMenu",
    "FactorCovarianceMeasureSettings",
    "FactorCovarianceReportSettings",
    "FactorCovarianceReportSettingsMenu",
    "IllegalPathError",
    "AsyncReportAccessorApi",
    "ReportAccessorSettings",
    "ReportAccessorApi",
    "AsyncReportPersister",
    "ReportPersister",
    "TimeSeriesBetaMeasureSettings",
    "TSBetaReportSettings",
    "RiskDatasetLoaderApi",
    "AsyncRiskDatasetLoaderApi",
    "AsyncCalendarApi",
    "AsyncCalendarLoaderApi",
    "CalendarApi",
    "CalendarLoaderApi",
    "CalendarSettings",
    "CalendarSettingsMenu",
    "UploadDataTypeFreshness",
    "UploadError",
    "UploaderApi",
    "AsyncUploaderApi",
    "UploadStagingResult",
    "UploadCommitResult",
    "MultiParserResult",
    "AsyncUploadersApi",
    "UploadersApi",
    "AsyncDataTypeUploaderApi",
    "DataTypeUploaderApi",
    "RiskDatasetApi",
    "AsyncRiskDatasetApi",
    "RiskDatasetSettings",
    "RiskDatasetSettingsMenu",
    "DatasetError",
    "RiskDatasetUpdateResult",
    "RiskDatasetExposureSettings",
    "RiskDatasetReferencedExposureSettings",
    "RiskDatasetUploadedExposureSettings",
    "RiskDatasetHuberRegressionExposureSettings",
    "RiskDatasetUnitExposureSettings",
    "RiskDatasetProperties",
    "RiskDatasetMetadata",
    "DatasetStatus",
    "GetModelMode",
    "FactorRiskModelMetadata",
    "hierarchy_to_df",
    "hierarchy_df_to_wide",
    "UploadSettings",
    "UploadSettingsMenu",
    "AsyncReportLoaderApiV2",
    "ReportLoaderApiV2",
    "AsyncReportApiV2",
    "ReportApiV2",
    "AsyncTypedReportAccessorApiV2",
    "TypedReportAccessorApiV2",
    "AsyncExposureReportApiV2",
    "ExposureReportApiV2",
    "ExposureReportSettingsV2",
    "AsyncFactorStatsReportApiV2",
    "FactorStatsReportApiV2",
    "FactorStatsReportSettingsV2",
    "AsyncUniverseCountReportApiV2",
    "UniverseCountReportApiV2",
    "UniverseCountReportSettingsV2",
    "AsyncFactorCovarianceReportApiV2",
    "FactorCovarianceReportApiV2",
    "FactorCovarianceReportSettingsV2",
    "AsyncPortfolioHoldingsReportApiV2",
    "PortfolioHoldingsReportApiV2",
    "PortfolioHoldingsReportSettingsV2",
    "AsyncFactorCovarianceReportAccessorV2",
    "AsyncFactorCovarianceReportAccessorImplV2",
    "AsyncUniverseCountReportAccessorV2",
    "AsyncUniverseCountReportAccessorImplV2",
    "AsyncExposureReportAccessorV2",
    "AsyncExposureReportAccessorImplV2",
    "AsyncFactorStatsReportAccessorV2",
    "AsyncFactorStatsReportAccessorImplV2",
    "ReportSettingsV2",
    "AsyncPortfolioHoldingsReportAccessorV2",
    "AsyncPortfolioHoldingsReportAccessorImplV2",
    "AnyReportSettings",
    "AsyncUniverseCountReportAccessorImplV2",
    "UniverseCountReportAccessorImplV2",
    "AsyncExposureReportAccessorImplV2",
    "ExposureReportAccessorImplV2",
    "AsyncFactorStatsReportAccessorImplV2",
    "FactorStatsReportAccessorImplV2",
    "AsyncFactorCovarianceReportAccessorImplV2",
    "FactorCovarianceReportAccessorImplV2",
    "AsyncPortfolioHoldingsReportAccessorImplV2",
    "PortfolioHoldingsReportAccessorImplV2",
    "AsyncPortfolioReturnsReportApiV2",
    "PortfolioReturnsReportApiV2",
    "PortfolioReturnsReportSettingsV2",
    "AsyncPortfolioReturnsReportAccessorV2",
    "AsyncPortfolioReturnsReportAccessorImplV2",
    "PortfolioReturnsReportAccessorImplV2",
    "AsyncReturnAttributionReportApiV2",
    "ReturnAttributionReportApiV2",
    "ReturnAttributionReportSettingsV2",
    "AsyncReturnAttributionReportAccessorV2",
    "AsyncReturnAttributionReportAccessorImplV2",
    "ReturnAttributionReportAccessorImplV2",
    "ACCESSOR_IMPLS",
    "ASYNC_ACCESSOR_IMPLS",
    "AsyncXSRReportApiV2",
    "XSRReportApiV2",
    "XSRReportSettingsV2",
    "AsyncXSRReportAccessorV2",
    "AsyncXSRReportAccessorImplV2",
    "XSRReportAccessorImplV2",
    "AsyncIdioReportApiV2",
    "IdioReportApiV2",
    "IdioReportSettingsV2",
    "AsyncIdioReportAccessorV2",
    "AsyncIdioReportAccessorImplV2",
    "IdioReportAccessorImplV2",
    "AsyncMcapExposureReportApiV2",
    "McapExposureReportApiV2",
    "McapExposureReportSettingsV2",
    "AsyncMcapExposureReportAccessorV2",
    "AsyncMcapExposureReportAccessorImplV2",
    "McapExposureReportAccessorImplV2",
    "AsyncWeightsReportApiV2",
    "WeightsReportApiV2",
    "WeightsReportSettingsV2",
    "AsyncWeightsReportAccessorV2",
    "AsyncWeightsReportAccessorImplV2",
    "WeightsReportAccessorImplV2",
    "PortfolioHoldingsScenario",
    "PortfolioHoldingsScenarioOverride",
    "ScenarioSettings",
    "AnyScenarioSettings",
    "AsyncFactorModelLoaderImpl",
    "AsyncFactorModelEngineImpl",
    "AsyncFactorModelImpl",
    "FactorModelLoaderImpl",
    "FactorModelEngineImpl",
    "FactorModelImpl",
    "AsyncFactorModelMetadataApi",
    "FactorModelMetadataApi",
]



import abc
import datetime as dt
from typing import Literal

import polars as pl

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.exposure_settings import (
    ExposureSettings,
    ExposureSettingsMenu,
)
from bayesline.api._src.equity.universe import AsyncUniverseApi, UniverseApi
from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.registry import AsyncRegistryBasedApi, RegistryBasedApi
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike, IdType


class ExposureApi(abc.ABC):
    """Abstract base class for exposure API operations."""

    @property
    @abc.abstractmethod
    def settings(self) -> ExposureSettings:
        """
        Get the settings used to create these exposures.

        Returns
        -------
        ExposureSettings
            The settings used to create these exposures.
        """
        ...

    @abc.abstractmethod
    def dates(
        self,
        universe: str | int | UniverseSettings | UniverseApi,
        *,
        range_only: bool = False,
    ) -> list[dt.date]:
        """
        Get the dates covered by the exposure data.

        Parameters
        ----------
        universe : str | int | UniverseSettings | UniverseApi
            The universe to use for the exposure calculation.
        range_only : bool, default=False
            If True, returns the first and last date only.

        Returns
        -------
        list[dt.date]
            List of all covered dates.
        """

    @abc.abstractmethod
    def coverage_stats(
        self,
        universe: str | int | UniverseSettings | UniverseApi,
        *,
        id_type: IdType | None = None,
        by: Literal["date", "asset"] = "date",
    ) -> pl.DataFrame:
        """
        Get coverage statistics for the exposure data.

        Parameters
        ----------
        universe : str | int | UniverseSettings | UniverseApi
            The universe to use for the exposure calculation.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        by : Literal["date", "asset"], default="date"
            The aggregation, either by date or by asset.

        Returns
        -------
        pl.DataFrame
            A dataframe with date as the first column, where the remaining columns
            names are the styles and substyles (concatenated with a dot). The values
            are the counts of the underlying data before it was imputed.
        """

    @abc.abstractmethod
    def coverage_stats_as_task(
        self,
        universe: str | int | UniverseSettings | UniverseApi,
        *,
        id_type: IdType | None = None,
        by: Literal["date", "asset"] = "date",
    ) -> Task[pl.DataFrame]:
        """
        Get coverage statistics as a task.

        Parameters
        ----------
        universe : str | int | UniverseSettings | UniverseApi
            The universe to use for the exposure calculation.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        by : Literal["date", "asset"], default="date"
            The aggregation, either by date or by asset.

        Returns
        -------
        Task[pl.DataFrame]
            A task that returns coverage statistics.
        """
        ...

    @abc.abstractmethod
    def get(
        self,
        universe: str | int | UniverseSettings | UniverseApi,
        *,
        standardize_universe: str | int | UniverseSettings | UniverseApi | None,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        filter_tradedays: bool = False,
    ) -> pl.DataFrame:
        """
        Get exposure data for the specified universe and date range.

        Parameters
        ----------
        universe : str | int | UniverseSettings | UniverseApi
            The universe to use for the exposure calculation.
        standardize_universe : str | int | UniverseSettings | UniverseApi | None
            The universe to use for the standardization of the exposure data. If None,
            the standardization is uses the entire universe.
        start : DateLike | None, default=None
            The start date of the universe to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        filter_tradedays : bool, default=False
            If True, only returns data for tradedays.

        Raises
        ------
        ValueError
            If the given id type is not supported or date range is invalid.

        Returns
        -------
        pl.DataFrame
            The data for the given date range with a where the date is the first column
            and the asset id is the second column. The remaining columns are the
            individual styles.
        """
        ...

    @abc.abstractmethod
    def get_as_task(
        self,
        universe: str | int | UniverseSettings | UniverseApi,
        *,
        standardize_universe: str | int | UniverseSettings | UniverseApi | None,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        filter_tradedays: bool = False,
    ) -> Task[pl.DataFrame]:
        """
        Get exposure data as a task.

        Parameters
        ----------
        universe : str | int | UniverseSettings | UniverseApi
            The universe to use for the exposure calculation.
        standardize_universe : str | int | UniverseSettings | UniverseApi | None
            The universe to use for the standardization of the exposure data. If None,
            the standardization is uses the entire universe.
        start : DateLike | None, default=None
            The start date of the universe to return, inclusive.
        end : DateLike | None, default=None
            The end date of the data to return, inclusive.
        id_type : IdType | None, default=None
            The id type to return asset ids in, e.g. `ticker`.
            The given id type must be supported by the linked universe.
        filter_tradedays : bool, default=False
            If True, only returns data for tradedays.

        Returns
        -------
        Task[pl.DataFrame]
            A task that returns exposure data.
        """
        ...


@docstrings_from_sync
class AsyncExposureApi(abc.ABC):

    @property
    @abc.abstractmethod
    def settings(self) -> ExposureSettings: ...  # noqa: D102

    @abc.abstractmethod
    async def dates(  # noqa: D102
        self,
        universe: str | int | UniverseSettings | AsyncUniverseApi,
        *,
        range_only: bool = False,
    ) -> list[dt.date]: ...

    @abc.abstractmethod
    async def coverage_stats(  # noqa: D102
        self,
        universe: str | int | UniverseSettings | AsyncUniverseApi,
        *,
        id_type: IdType | None = None,
        by: Literal["date", "asset"] = "date",
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def coverage_stats_as_task(  # noqa: D102
        self,
        universe: str | int | UniverseSettings | AsyncUniverseApi,
        *,
        id_type: IdType | None = None,
        by: Literal["date", "asset"] = "date",
    ) -> AsyncTask[pl.DataFrame]: ...

    @abc.abstractmethod
    async def get(  # noqa: D102
        self,
        universe: str | int | UniverseSettings | AsyncUniverseApi,
        *,
        standardize_universe: str | int | UniverseSettings | AsyncUniverseApi | None,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        filter_tradedays: bool = False,
    ) -> pl.DataFrame: ...

    @abc.abstractmethod
    async def get_as_task(  # noqa: D102
        self,
        universe: str | int | UniverseSettings | AsyncUniverseApi,
        *,
        standardize_universe: str | int | UniverseSettings | AsyncUniverseApi | None,
        start: DateLike | None = None,
        end: DateLike | None = None,
        id_type: IdType | None = None,
        filter_tradedays: bool = False,
    ) -> AsyncTask[pl.DataFrame]: ...


class ExposureLoaderApi(
    RegistryBasedApi[ExposureSettings, ExposureSettingsMenu, ExposureApi],
):
    """Registry-based API for loading exposure data."""

    ...


@docstrings_from_sync
class AsyncExposureLoaderApi(
    AsyncRegistryBasedApi[ExposureSettings, ExposureSettingsMenu, AsyncExposureApi],
): ...



import abc
from typing import Any, Literal

import polars as pl
from pydantic import Field

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.settings import (
    FactorCovarianceSettings,
    ReportSettings,
)
from bayesline.api._src.equity.riskmodels_settings import FactorRiskModelSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike


class FactorCovarianceReportSettings(ReportSettings):
    """Settings for a factor covariance report."""

    report_type: Literal["factor_covariance"] = Field(
        default="factor_covariance", description="Report type discriminator"
    )
    factor_model_settings: FactorRiskModelSettings = Field()
    factor_cov_settings: FactorCovarianceSettings = Field(
        default_factory=FactorCovarianceSettings,
        description="Settings for the factor covariance matrix forecasts.",
    )


class FactorCovarianceReportAccessor(TypedReportAccessorApi):
    """Specific accessor for a exposure report."""

    @abc.abstractmethod
    def get_covariance(self) -> pl.DataFrame:
        """
        Get the time-series of factor covariances for the factors.

        Returns
        -------
        pl.DataFrame
            A dataframe with a `date` and `factor` columns and columns for each factor
            with their respective factor covariances.
        """

    @abc.abstractmethod
    def get_correlation(self) -> pl.DataFrame:
        """
        Get the time-series of factor correlations for the factors.

        Returns
        -------
        pl.DataFrame
            A dataframe with a `date` and `factor` columns and columns for each factor
            with their respective factor correlations.
        """


@docstrings_from_sync
class AsyncFactorCovarianceReportAccessor(AsyncTypedReportAccessorApi):

    @abc.abstractmethod
    async def get_covariance(self) -> pl.DataFrame: ...  # noqa: D102

    @abc.abstractmethod
    async def get_correlation(self) -> pl.DataFrame: ...  # noqa: D102


class FactorCovarianceReportApi(
    ReportApi[
        [DateLike, DateLike],
        FactorCovarianceReportAccessor,
        FactorCovarianceReportSettings,
    ]
):
    """API for a exposure report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> FactorCovarianceReportAccessor | None:
        """
        Get the cached factor covariance report accessor API if it is available.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        FactorCovarianceReportAccessor | None
            The cached factor covariance report accessor API, or None if the report is
            not cached.
        """

    @abc.abstractmethod
    def calculate(
        self,
        start_date: DateLike,
        end_date: DateLike,
        **kwargs: Any,
    ) -> FactorCovarianceReportAccessor:
        """
        Calculate the factor covariance report.

        Parameters
        ----------
        start_date: DateLike
            The start date of the report.
        end_date: DateLike
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        FactorCovarianceReportAccessor
            The exposure report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[FactorCovarianceReportAccessor]:
        """
        Calculate the factor covariance report as a task.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[FactorCovarianceReportAccessor]
            The task for the factor covariance report accessor.
        """


@docstrings_from_sync
class AsyncFactorCovarianceReportApi(
    AsyncReportApi[
        [DateLike | None, DateLike | None],
        AsyncFactorCovarianceReportAccessor,
        FactorCovarianceReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncFactorCovarianceReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncFactorCovarianceReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncFactorCovarianceReportAccessor]: ...


class FactorCovarianceReportAccessorImpl(FactorCovarianceReportAccessor):  # noqa: D101

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor

    def get_covariance(self) -> pl.DataFrame:  # noqa: D102
        df = self.accessor.get_data(
            [],
            expand=("date", "factor"),
            value_cols=("covariance",),
            pivot_cols=("factor_col",),
        )
        return df.rename(lambda x: x.removesuffix("^covariance"))

    def get_correlation(self) -> pl.DataFrame:  # noqa: D102
        df = self.accessor.get_data(
            [],
            expand=("date", "factor"),
            value_cols=("correlation",),
            pivot_cols=("factor_col",),
        )
        return df.rename(lambda x: x.removesuffix("^correlation"))


class AsyncFactorCovarianceReportAccessorImpl(  # noqa: D101
    AsyncFactorCovarianceReportAccessor
):

    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor

    async def get_covariance(self) -> pl.DataFrame:  # noqa: D102
        df = await self.accessor.get_data(
            [],
            expand=("date", "factor"),
            value_cols=("covariance",),
            pivot_cols=("factor_col",),
        )
        return df.rename(lambda x: x.removesuffix("^covariance"))

    async def get_correlation(self) -> pl.DataFrame:  # noqa: D102
        df = await self.accessor.get_data(
            [],
            expand=("date", "factor"),
            value_cols=("correlation",),
            pivot_cols=("factor_col",),
        )
        return df.rename(lambda x: x.removesuffix("^correlation"))



"""Bayesline API package for equity data access and analysis."""

__version__ = "0.12.1"

from bayesline.api._src.acl_models import (
    ACLEntry,
    ACLError,
    ACLValidationError,
    CreateACLEntry,
    CreateGroupMembership,
    CreateUserGroup,
    CreateUserRole,
    GroupCreationError,
    GroupDeletionError,
    GroupMembership,
    GroupMembershipError,
    GroupUpdateError,
    PermissionDeniedError,
    PermissionGrantError,
    PermissionLevel,
    PrincipalNotFoundError,
    PrincipalType,
    ResourceNotFoundError,
    ResourcePermissionSummary,
    ResourceType,
    RoleAssignmentError,
    RoleType,
    UpdateACLEntry,
    UpdateUserGroup,
    UserGroup,
    UserPermissionsSummary,
    UserRole,
)
from bayesline.api._src.api import AsyncBayeslineApi, BayeslineApi
from bayesline.api._src.maintenance import (
    AsyncIncidentsServiceApi,
    IncidentsServiceApi,
    IncidentSummary,
    IncidentSummaryItem,
)
from bayesline.api._src.migration import (
    migrate_raw_json,
    stamp_version,
    validate_migration_chain,
)
from bayesline.api._src.permissions import AsyncPermissionsApi, PermissionsApi
from bayesline.api._src.registry import (
    AsyncReadOnlyRegistry,
    AsyncRegistry,
    AsyncSettingsRegistry,
    AsyncSettingsResolver,
    EmptySettingsMenu,
    InvalidSettingsError,
    Mode,
    RawSettings,
    ReadOnlyRegistry,
    Registry,
    Settings,
    SettingsMenu,
    SettingsMenuType,
    SettingsMetaData,
    SettingsRegistry,
    SettingsType,
    SettingsTypeMetaData,
)
from bayesline.api._src.settings import (
    AsyncSettingsRegistry as NewAsyncSettingsRegistry,
)
from bayesline.api._src.settings import (
    AsyncTypedSettingsRegistry,
    CouldNotDeleteSettingsError,
    CouldNotPermissionSettingsError,
    CouldNotSaveSettingsError,
    DeleteItemResult,
    DeleteResult,
    ItemResult,
    ItemStatus,
    PermissionItemResult,
    PermissionResult,
    SaveItemResult,
    SaveResult,
    SettingInvalid,
    SettingNotFound,
    SettingNotPermissioned,
    SettingResult,
    SettingsIdentifiers,
)
from bayesline.api._src.settings import SettingsRegistry as NewSettingsRegistry
from bayesline.api._src.settings import (
    SettingsRegistryError,
    SettingsRegistryResult,
    SettingsTransferObject,
    SettingValid,
    TypedSettingsRegistry,
)
from bayesline.api._src.settings_types import (
    SETTINGS_TYPES,
)
from bayesline.api._src.tasks import (
    AsyncTask,
    AsyncTasksApi,
    Task,
    TaskError,
    TaskProgress,
    TaskResponse,
    TasksApi,
    TaskState,
)

__all__ = [
    "SettingsType",
    "SettingsMenuType",
    "types",
    "SettingsMenu",
    "EmptySettingsMenu",
    "BayeslineApi",
    "AsyncBayeslineApi",
    "SettingsRegistry",
    "AsyncSettingsRegistry",
    "Registry",
    "AsyncRegistry",
    "SettingsMetaData",
    "PermissionsApi",
    "AsyncPermissionsApi",
    "Settings",
    "SettingsTypeMetaData",
    "Mode",
    "RawSettings",
    "InvalidSettingsError",
    "AsyncSettingsResolver",
    "ReadOnlyRegistry",
    "AsyncReadOnlyRegistry",
    "IncidentsServiceApi",
    "IncidentSummary",
    "IncidentSummaryItem",
    "AsyncIncidentsServiceApi",
    "TasksApi",
    "AsyncTasksApi",
    "TaskProgress",
    "TaskState",
    "Task",
    "AsyncTask",
    "TaskError",
    "TaskResponse",
    "SettingsIdentifiers",
    "TypedSettingsRegistry",
    "NewSettingsRegistry",
    "SETTINGS_TYPES",
    "NewAsyncSettingsRegistry",
    "AsyncTypedSettingsRegistry",
    "ACLEntry",
    "ACLError",
    "ACLValidationError",
    "CreateACLEntry",
    "CreateGroupMembership",
    "CreateUserGroup",
    "CreateUserRole",
    "GroupCreationError",
    "GroupDeletionError",
    "GroupMembership",
    "GroupMembershipError",
    "GroupUpdateError",
    "PermissionDeniedError",
    "PermissionGrantError",
    "PermissionLevel",
    "PrincipalNotFoundError",
    "PrincipalType",
    "ResourceNotFoundError",
    "ResourcePermissionSummary",
    "ResourceType",
    "RoleAssignmentError",
    "RoleType",
    "UpdateACLEntry",
    "UpdateUserGroup",
    "UserGroup",
    "UserPermissionsSummary",
    "UserRole",
    "ACLEntry",
    "ACLError",
    "ACLValidationError",
    "CreateACLEntry",
    "CreateGroupMembership",
    "CreateUserGroup",
    "CreateUserRole",
    "GroupCreationError",
    "GroupDeletionError",
    "GroupMembership",
    "GroupMembershipError",
    "GroupUpdateError",
    "PermissionDeniedError",
    "PermissionGrantError",
    "PermissionLevel",
    "PrincipalNotFoundError",
    "PrincipalType",
    "ResourceNotFoundError",
    "ResourcePermissionSummary",
    "ResourceType",
    "RoleAssignmentError",
    "RoleType",
    "UpdateACLEntry",
    "UpdateUserGroup",
    "UserGroup",
    "UserPermissionsSummary",
    "UserRole",
    "CouldNotSaveSettingsError",
    "SettingResult",
    "SettingNotFound",
    "SettingInvalid",
    "SettingValid",
    "SettingNotPermissioned",
    "SettingsTransferObject",
    "CouldNotDeleteSettingsError",
    "CouldNotPermissionSettingsError",
    "ItemResult",
    "ItemStatus",
    "SettingsRegistryResult",
    "SettingsRegistryError",
    "SaveResult",
    "DeleteResult",
    "PermissionResult",
    "SaveItemResult",
    "DeleteItemResult",
    "PermissionItemResult",
    "migrate_raw_json",
    "stamp_version",
    "validate_migration_chain",
]



import abc
import enum
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, Field

from bayesline.api._src._utils import docstrings_from_sync

T = TypeVar("T", covariant=True)


class TaskResponse(BaseModel):
    """Response model for task operations."""

    task_id: str = Field(description="The unique identifier for the task.")
    extra: dict[str, Any] = Field(
        default_factory=dict, description="Additional metadata or context for the task."
    )


class TaskState(str, enum.Enum):
    """The state of a task."""

    CANCELLED = "CANCELLED"
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    RETRY = "RETRY"


class TaskError(Exception):
    """Exception raised when a task operation fails."""


class TaskProgress(BaseModel):
    """The progress of a task.

    Attributes
    ----------
    state : TaskState
        Current state of the task.
    last_progress : int
        Last reported progress percentage (0-100).
    last_message : str
        Last status message from the task.
    last_context : str, default=''
        Additional context information about the task.
    queue_position : int | None, default=None
        Number of tasks ahead in the queue (0 = next to run).
        Only populated when state is QUEUED.
    """

    state: TaskState
    last_progress: int
    last_message: str
    last_context: str = ""
    queue_position: int | None = None


class Task(abc.ABC, Generic[T]):
    """
    Abstract base class representing a task that produces a result of type T.

    A task has a unique ID and can be in various states (queued, running, completed, etc).
    The task's progress can be monitored and its result retrieved once complete.
    """

    @property
    @abc.abstractmethod
    def task_id(self) -> str:
        """The unique identifier for this task.

        Returns
        -------
        str
            The unique identifier for this task.
        """

    @abc.abstractmethod
    def get_progress(self) -> TaskProgress:
        """
        Get the current progress status of the task.

        Returns
        -------
        TaskProgress
            The current progress status, containing information about
            completion percentage and status message.
        """

    @abc.abstractmethod
    def is_ready(self) -> bool:
        """
        Check if the task has completed and the result is ready.

        Returns
        -------
        bool
            True if the task is complete and result can be retrieved, False otherwise.
        """

    @abc.abstractmethod
    def get_result(self) -> T:
        """
        Get the result of the completed task.

        Returns
        -------
        T
            The result of the task.

        Raises
        ------
        TaskError
            If the task failed or is not yet complete.
        """

    @abc.abstractmethod
    def wait_result(self, timeout: float = -1.0, check_interval: float = 0.5) -> T:
        """
        Wait for the task to complete and return its result.

        Parameters
        ----------
        timeout : float, default=-1.0
            Maximum time in seconds to wait for completion.
            A negative value means wait indefinitely.
        check_interval : float, default=0.5
            Time in seconds between status checks.

        Returns
        -------
        T
            The result of the task.

        Raises
        ------
        TaskError
            If the task failed or timed out.
        """

    @abc.abstractmethod
    def wait_ready(self, timeout: float = -1.0, check_interval: float = 1.0) -> None:
        """
        Wait for the task to be ready.

        Parameters
        ----------
        timeout : float, default=-1.0
            Maximum time in seconds to wait for completion.
            A negative value means wait indefinitely.
        check_interval : float, default=1.0
            Time in seconds between status checks.

        Raises
        ------
        TaskError
            If the task failed or timed out.
        """

    @abc.abstractmethod
    def cancel(self) -> None:
        """Cancel the task."""


@docstrings_from_sync
class AsyncTask(abc.ABC, Generic[T]):

    @property
    @abc.abstractmethod
    def task_id(self) -> str: ...  # noqa: D102

    @abc.abstractmethod
    async def get_progress(self) -> TaskProgress: ...  # noqa: D102

    @abc.abstractmethod
    async def is_ready(self) -> bool: ...  # noqa: D102

    @abc.abstractmethod
    async def get_result(self) -> T: ...  # noqa: D102

    @abc.abstractmethod
    async def wait_result(  # noqa: D102
        self, timeout: float = -1.0, check_interval: float = 1.0
    ) -> T: ...

    @abc.abstractmethod
    async def wait_ready(  # noqa: D102
        self, timeout: float = -1.0, check_interval: float = 1.0
    ) -> None: ...

    @abc.abstractmethod
    async def cancel(self) -> None: ...  # noqa: D102


class TasksApi(abc.ABC):
    """Abstract base class for synchronous task API operations."""

    @abc.abstractmethod
    def get_task_progress(self, task_id: str) -> TaskProgress:
        """
        Get the progress of a task if it exists.

        Parameters
        ----------
        task_id : str
            The ID of the task.

        Returns
        -------
        TaskProgress
            The progress of the task.

        Raises
        ------
        KeyError
            If the task does not exist.
        """


@docstrings_from_sync
class AsyncTasksApi(abc.ABC):

    @abc.abstractmethod
    async def get_task_progress(self, task_id: str) -> TaskProgress: ...  # noqa: D102



from bayesline.api._src.equity.report.covariance import (
    AsyncFactorCovarianceReportAccessorImpl,
    FactorCovarianceReportAccessorImpl,
    FactorCovarianceReportSettings,
)
from bayesline.api._src.equity.report.exposure import (
    AsyncExposureReportAccessorImpl,
    ExposureReportAccessorImpl,
    ExposureReportSettings,
)
from bayesline.api._src.equity.report.factor_stats import (
    AsyncFactorStatsReportAccessorImpl,
    FactorStatsReportAccessorImpl,
    FactorStatsReportSettings,
)
from bayesline.api._src.equity.report.idio import (
    AsyncIdioReportAccessorImpl,
    IdioReportAccessorImpl,
    IdioReportSettings,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    AsyncMcapExposureReportAccessorImpl,
    McapExposureReportAccessorImpl,
    McapExposureReportSettings,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    AsyncPortfolioHoldingsReportAccessorImpl,
    PortfolioHoldingsReportAccessorImpl,
    PortfolioHoldingsReportSettings,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    AsyncPortfolioReturnsReportAccessorImpl,
    PortfolioReturnsReportAccessorImpl,
    PortfolioReturnsReportSettings,
)
from bayesline.api._src.equity.report.return_attribution import (
    AsyncReturnAttributionReportAccessorImpl,
    ReturnAttributionReportAccessorImpl,
    ReturnAttributionReportSettings,
)
from bayesline.api._src.equity.report.universe import (
    AsyncUniverseCountReportAccessorImpl,
    UniverseCountReportAccessorImpl,
    UniverseCountReportSettings,
)
from bayesline.api._src.equity.report.weights import (
    AsyncWeightsReportAccessorImpl,
    WeightsReportAccessorImpl,
    WeightsReportSettings,
)
from bayesline.api._src.equity.report.xsr import (
    AsyncXSRReportAccessorImpl,
    XSRReportAccessorImpl,
    XSRReportSettings,
)

ACCESSOR_IMPLS = {
    UniverseCountReportSettings: UniverseCountReportAccessorImpl,
    ExposureReportSettings: ExposureReportAccessorImpl,
    FactorStatsReportSettings: FactorStatsReportAccessorImpl,
    FactorCovarianceReportSettings: FactorCovarianceReportAccessorImpl,
    PortfolioHoldingsReportSettings: PortfolioHoldingsReportAccessorImpl,
    PortfolioReturnsReportSettings: PortfolioReturnsReportAccessorImpl,
    ReturnAttributionReportSettings: ReturnAttributionReportAccessorImpl,
    XSRReportSettings: XSRReportAccessorImpl,
    IdioReportSettings: IdioReportAccessorImpl,
    McapExposureReportSettings: McapExposureReportAccessorImpl,
    WeightsReportSettings: WeightsReportAccessorImpl,
}
ASYNC_ACCESSOR_IMPLS = {
    UniverseCountReportSettings: AsyncUniverseCountReportAccessorImpl,
    ExposureReportSettings: AsyncExposureReportAccessorImpl,
    FactorStatsReportSettings: AsyncFactorStatsReportAccessorImpl,
    FactorCovarianceReportSettings: AsyncFactorCovarianceReportAccessorImpl,
    PortfolioHoldingsReportSettings: AsyncPortfolioHoldingsReportAccessorImpl,
    PortfolioReturnsReportSettings: AsyncPortfolioReturnsReportAccessorImpl,
    ReturnAttributionReportSettings: AsyncReturnAttributionReportAccessorImpl,
    XSRReportSettings: AsyncXSRReportAccessorImpl,
    IdioReportSettings: AsyncIdioReportAccessorImpl,
    McapExposureReportSettings: AsyncMcapExposureReportAccessorImpl,
    WeightsReportSettings: AsyncWeightsReportAccessorImpl,
}



"""Shared status types for risk datasets and risk models."""

from typing import Literal, TypeAlias

DatasetStatus: TypeAlias = Literal[
    "ready", "data_stale", "needs_update", "not_ready", "corrupt"
]
"""Status type for risk datasets and risk models.

Status definitions:
- ``ready``: Data is up-to-date and ready to use.
- ``data_stale``: Data exists but underlying source data has been updated.
- ``needs_update``: No local data exists, or data exists but settings have changed.
- ``not_ready``: Reference dataset is not in a usable state.
- ``corrupt``: Settings are invalid or data is corrupted.
"""



import abc
from typing import Annotated, Union, overload

from pydantic import Discriminator

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import (
    AsyncReportApi,
    ReportApi,
    S,
)
from bayesline.api._src.equity.report.covariance import (
    AsyncFactorCovarianceReportApi,
    FactorCovarianceReportApi,
    FactorCovarianceReportSettings,
)
from bayesline.api._src.equity.report.exposure import (
    AsyncExposureReportApi,
    ExposureReportApi,
    ExposureReportSettings,
)
from bayesline.api._src.equity.report.factor_stats import (
    AsyncFactorStatsReportApi,
    FactorStatsReportApi,
    FactorStatsReportSettings,
)
from bayesline.api._src.equity.report.idio import (
    AsyncIdioReportApi,
    IdioReportApi,
    IdioReportSettings,
)
from bayesline.api._src.equity.report.mcap_exposure import (
    AsyncMcapExposureReportApi,
    McapExposureReportApi,
    McapExposureReportSettings,
)
from bayesline.api._src.equity.report.portfolio_holdings import (
    AsyncPortfolioHoldingsReportApi,
    PortfolioHoldingsReportApi,
    PortfolioHoldingsReportSettings,
)
from bayesline.api._src.equity.report.portfolio_returns import (
    AsyncPortfolioReturnsReportApi,
    PortfolioReturnsReportApi,
    PortfolioReturnsReportSettings,
)
from bayesline.api._src.equity.report.return_attribution import (
    AsyncReturnAttributionReportApi,
    ReturnAttributionReportApi,
    ReturnAttributionReportSettings,
)
from bayesline.api._src.equity.report.universe import (
    AsyncUniverseCountReportApi,
    UniverseCountReportApi,
    UniverseCountReportSettings,
)
from bayesline.api._src.equity.report.weights import (
    AsyncWeightsReportApi,
    WeightsReportApi,
    WeightsReportSettings,
)
from bayesline.api._src.equity.report.xsr import (
    AsyncXSRReportApi,
    XSRReportApi,
    XSRReportSettings,
)

# Discriminated union of all report settings for polymorphic deserialization
AnyReportSettings = Annotated[
    Union[
        UniverseCountReportSettings,
        ExposureReportSettings,
        FactorStatsReportSettings,
        FactorCovarianceReportSettings,
        PortfolioHoldingsReportSettings,
        PortfolioReturnsReportSettings,
        ReturnAttributionReportSettings,
        XSRReportSettings,
        IdioReportSettings,
        McapExposureReportSettings,
        WeightsReportSettings,
    ],
    Discriminator("report_type"),
]


class ReportLoaderApi(abc.ABC):
    """The main interface for loading different types of reports."""

    @overload
    def load(self, settings: UniverseCountReportSettings) -> UniverseCountReportApi: ...

    @overload
    def load(self, settings: ExposureReportSettings) -> ExposureReportApi: ...

    @overload
    def load(self, settings: FactorStatsReportSettings) -> FactorStatsReportApi: ...

    @overload
    def load(
        self, settings: FactorCovarianceReportSettings
    ) -> FactorCovarianceReportApi: ...

    @overload
    def load(
        self, settings: PortfolioHoldingsReportSettings
    ) -> PortfolioHoldingsReportApi: ...

    @overload
    def load(
        self, settings: PortfolioReturnsReportSettings
    ) -> PortfolioReturnsReportApi: ...

    @overload
    def load(
        self, settings: ReturnAttributionReportSettings
    ) -> ReturnAttributionReportApi: ...

    @overload
    def load(self, settings: IdioReportSettings) -> IdioReportApi: ...

    @overload
    def load(self, settings: XSRReportSettings) -> XSRReportApi: ...

    @overload
    def load(self, settings: McapExposureReportSettings) -> McapExposureReportApi: ...

    @overload
    def load(self, settings: WeightsReportSettings) -> WeightsReportApi: ...

    @abc.abstractmethod
    def load(self, settings: S) -> ReportApi[..., TypedReportAccessorApi, S]:
        """
        Load a report using the specified settings.

        Parameters
        ----------
        settings: S
            The settings to use to load the report.

        Returns
        -------
        ReportApi[..., TypedReportAccessorApi, S]
            The loaded report API.
        """

    @abc.abstractmethod
    def load_accessor(self, identifier: str) -> TypedReportAccessorApi:
        """
        Load a report accessor from an identifier.

        This accessor is obtained from an existing accessor after a report has been
        calculated. The identifier is ephemeral, and should only be used for short-
        lived sessions, and not persisted over time.

        Parameters
        ----------
        identifier: str
            The short-lived ephemeral identifier to load the report accessor from.

        Returns
        -------
        TypedReportAccessorApi
            The loaded report accessor.

        Raises
        ------
        KeyError
            If the identifier is not found. This can happen if the identifier has
            expired, belongs to a different session, or was never created.
        """


@docstrings_from_sync
class AsyncReportLoaderApi(abc.ABC):

    @overload
    async def load(  # noqa: D102
        self, settings: UniverseCountReportSettings
    ) -> AsyncUniverseCountReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: ExposureReportSettings
    ) -> AsyncExposureReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: FactorStatsReportSettings
    ) -> AsyncFactorStatsReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: FactorCovarianceReportSettings
    ) -> AsyncFactorCovarianceReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: PortfolioHoldingsReportSettings
    ) -> AsyncPortfolioHoldingsReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: PortfolioReturnsReportSettings
    ) -> AsyncPortfolioReturnsReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: ReturnAttributionReportSettings
    ) -> AsyncReturnAttributionReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: IdioReportSettings
    ) -> AsyncIdioReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: XSRReportSettings
    ) -> AsyncXSRReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: McapExposureReportSettings
    ) -> AsyncMcapExposureReportApi: ...

    @overload
    async def load(  # noqa: D102
        self, settings: WeightsReportSettings
    ) -> AsyncWeightsReportApi: ...

    @abc.abstractmethod
    async def load(  # noqa: D102
        self, settings: S
    ) -> AsyncReportApi[..., AsyncTypedReportAccessorApi, S]: ...

    @abc.abstractmethod
    async def load_accessor(  # noqa: D102
        self, identifier: str
    ) -> AsyncTypedReportAccessorApi: ...



from typing import Literal

from pydantic import BaseModel, Field

from bayesline.api._src.registry import Settings, SettingsMenu


class ReportSettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """Defines the available report types and settings to create a report."""

    pass


class ReportSettings(Settings[ReportSettingsMenu]):
    """Concrete settings for a report.

    Subclasses must override `report_type` with a Literal type for polymorphic
    deserialization via the discriminated union `AnyReportSettings`.
    """

    report_type: str = Field(description="Discriminator field for report type")

    @property
    def menu_type(self) -> type[ReportSettingsMenu]:
        """The menu type for this settings object."""
        return ReportSettingsMenu


class FactorCovarianceSettings(BaseModel, frozen=True, extra="forbid"):
    """Settings for factor covariance matrix forecasts."""

    halflife_vol: float = Field(
        default=60.0, description="The half-life for the factor volatility."
    )
    halflife_cor: float = Field(
        default=120.0, description="The half-life for the factor correlation."
    )
    halflife_vra: float | None = Field(
        default=None,
        description=(
            "The half-life for the factor volatility regime adjustment. If None, no "
            "adjustment is applied."
        ),
    )
    nw_lags_vol: int = Field(
        default=0, description="The Newey-West lags of the factor volatility."
    )
    nw_lags_vol_halflife_override: float | None = Field(
        default=None,
        description=(
            "The Newey-West lags override for the factor volatility for lagged returns."
        ),
    )
    nw_lags_cor: int = Field(
        default=0, description="The Newey-West lags of the factor correlation."
    )
    shrink_cor_method: Literal["LIS", "QIS", "GIS"] | None = Field(
        default=None,
        description=(
            "The method to use for the shrinkage of the factor correlation. If None, "
            "don't shrink."
        ),
    )
    shrink_cor_length: int | None = Field(
        default=None,
        description=(
            "If a shrinkage method is provided, the length of the shrinkage for the "
            "factor correlation."
        ),
    )
    combine_standardized: bool = Field(
        default=False,
        description=(
            "If True, the correlation matrix is combined with an estimate based on the "
            "standardized returns. The halflife_factor_vra is used for the weights."
        ),
    )


class IdioVolSettings(BaseModel, frozen=True, extra="forbid"):
    """Settings for idiosyncratic volatility matrix forecasts."""

    halflife: float = Field(
        default=60.0, description="The half-life for the idiosyncratic volatility."
    )
    nw_lags: int = Field(
        default=0, description="The Newey-West lags of the idiosyncratic volatility."
    )
    nw_lags_halflife_override: float | None = Field(
        default=None,
        description=(
            "The Newey-West lags override for the idiosyncratic volatility for lagged "
            "returns."
        ),
    )
    clip_lb: float | None = Field(
        default=-0.3,
        description=(
            "The lower bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    clip_ub: float | None = Field(
        default=None,
        description=(
            "The upper bound to clip the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no clipping is applied."
        ),
    )
    void_lb: float | None = Field(
        default=None,
        description=(
            "The lower bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no removal is applied. Removal of "
            "returns is applied before clipping."
        ),
    )
    void_ub: float | None = Field(
        default=0.5,
        description=(
            "The upper bound to remove the idiosyncratic return in order to compute "
            "the idiosyncratic volatility. If None, no removal is applied. Removal of "
            "returns is applied before clipping."
        ),
    )



from collections.abc import Sequence
from typing import Annotated, Any

from pydantic import AfterValidator, Field, ValidationInfo, field_validator

from bayesline.api._src.registry import Settings, SettingsMenu


def require_sorted_unique(v: Any) -> Any:
    """
    Validate that the given list is sorted and unique.

    Parameters
    ----------
    v : Any
        The list to validate.

    Returns
    -------
    Any
        The validated list.

    Raises
    ------
    ValueError
        If the list is not unique or not sorted.
    """
    if len(v) != len(set(v)):
        raise ValueError("The list must be unique.")
    if v != sorted(v):
        raise ValueError("The list must be sorted.")
    return v


class CalendarSettingsMenu(SettingsMenu, frozen=True, extra="forbid"):
    """Contain the available settings that can be used for the calendar settings."""

    exchanges: Annotated[list[str], AfterValidator(require_sorted_unique)] = Field(
        description="""
        A list of mic codes of all exchanges. Must be sorted and unique.
        """,
    )

    def describe(self) -> str:
        """
        Describe the calendar settings menu.

        Returns
        -------
        str
            Description of the settings menu.
        """
        return f"Exchanges (MIC): {', '.join(self.exchanges)}"


class CalendarSettings(Settings[CalendarSettingsMenu]):
    """Define the settings for the calendar."""

    dataset: str | None = Field(
        default=None,
        description=(
            "The name of the underlying dataset to use. If none is given then the "
            "configured default dataset is used."
        ),
        examples=["Bayesline-Global"],
    )

    filters: list[list[str]] = Field(
        default=[["XNYS"]],
        min_length=1,
        description=(
            "The filters to apply. Each filter is a list of exchange MIC codes. The "
            "outer list will be treated as an OR conditions, while the inner lists "
            "will be treated as an AND conditions. For example, `[['A', 'B'], ['C']]` "
            "means that the holidays are the days where either A and B are both "
            "holidays, or C is a holiday."
        ),
        examples=[[["XNYS"]], [["XNYS", "XNAS"]], [["XNYS"], ["XNAS"]]],
    )

    def get_references(self) -> Sequence[str | int]:
        """
        Get references for this settings object.

        Returns
        -------
        Sequence[str | int]
            A sequence of references (strings or integers) for this settings object.
        """
        result: list[str | int] = []
        if self.dataset is not None:
            result.append(self.dataset)
        return result

    @property
    def menu_type(self) -> type[CalendarSettingsMenu]:  # noqa: D102
        return CalendarSettingsMenu

    @field_validator("filters", mode="after")
    @classmethod
    def validate_filters(
        cls, value: list[list[str]], info: ValidationInfo
    ) -> list[list[str]]:
        if isinstance(menu := info.context, CalendarSettingsMenu):
            not_found = set()
            for filter_or in value:
                for filter_and in filter_or:
                    if filter_and not in menu.exchanges:
                        not_found.add(filter_and)
            if not_found:
                raise ValueError(
                    f"""
                    The following exchanges do not exist: {', '.join(not_found)}.
                    Only {', '.join(menu.exchanges)} exist.
                    """,
                )
        return value

    def describe(self, menu: CalendarSettingsMenu) -> str:
        """Describe the calendar settings.

        Parameters
        ----------
        menu : CalendarSettingsMenu
            The settings menu to use to describe the settings.

        Returns
        -------
        str
            A description of the calendar settings.
        """
        del menu  # not used
        filter_str = " or ".join(
            (
                f"({' and '.join(and_terms)})"
                if len(self.filters) > 1 and len(and_terms) > 1
                else f"{' and '.join(and_terms)}"
            )
            for and_terms in self.filters
        )
        return f"Holidays where holidays on {filter_str}."



import abc
from typing import Any, Literal, Self

from pydantic import Field, model_validator

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.calendar_settings import CalendarSettings
from bayesline.api._src.equity.portfoliohierarchy_settings import (
    PortfolioHierarchySettings,
)
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.scenarios import PortfolioHoldingsScenario
from bayesline.api._src.equity.report.settings import ReportSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike


class PortfolioHoldingsReportSettings(ReportSettings):
    """Settings for a portfolio holdings report."""

    report_type: Literal["portfolio_holdings"] = Field(
        default="portfolio_holdings", description="Report type discriminator"
    )
    portfolio_hierarchy_settings: PortfolioHierarchySettings = Field()
    date_aggregation: Literal["last"] = Field("last")
    calendar: CalendarSettings = Field(default_factory=CalendarSettings)
    currency: str | None = Field(None)
    unpack: bool = Field(False)
    forward_fill: bool = Field(False)

    @model_validator(mode="after")
    def validate_currency_unpack_consistency(self) -> Self:
        if self.currency is None and self.unpack:
            raise ValueError("currency is required when unpack is True")
        return self


class PortfolioHoldingsReportAccessor(TypedReportAccessorApi):
    """Specific accessor for a portfolio holdings report."""

    scenario_types = [PortfolioHoldingsScenario]


@docstrings_from_sync
class AsyncPortfolioHoldingsReportAccessor(AsyncTypedReportAccessorApi):
    scenario_types = [PortfolioHoldingsScenario]


class PortfolioHoldingsReportApi(
    ReportApi[
        [list[str] | None, DateLike | None, DateLike | None],
        PortfolioHoldingsReportAccessor,
        PortfolioHoldingsReportSettings,
    ]
):
    """API for a universe count report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> PortfolioHoldingsReportAccessor | None:
        """
        Get the cached portfolio holdings report accessor API if it is available.

        Parameters
        ----------
        portfolio_names: list[str] | None
            The names of the portfolios to calculate the report for. If None, all
            portfolios will be calculated. For nested portfolios, the names must be
            given in the format "portfolio:held_portfolio", with multiple separators
            indicating deeper nesting. If nested portfolios are given, unpack must be
            True in the report settings.
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        PortfolioHoldingsReportAccessor | None
            The cached portfolio holdings report accessor API, or None if the report is
            not cached.
        """

    @abc.abstractmethod
    def calculate(
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> PortfolioHoldingsReportAccessor:
        """
        Calculate the portfolio holdings report.

        Parameters
        ----------
        portfolio_names: list[str] | None
            The names of the portfolios to calculate the report for. If None, all
            portfolios will be calculated. For nested portfolios, the names must be
            given in the format "portfolio:held_portfolio", with multiple separators
            indicating deeper nesting. If nested portfolios are given, unpack must be
            True in the report settings.
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        PortfolioHoldingsReportAccessor
            The portfolio holdings report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[PortfolioHoldingsReportAccessor]:
        """
        Calculate the portfolio holdings report as a task.

        Parameters
        ----------
        portfolio_names: list[str] | None
            The names of the portfolios to calculate the report for. If None, all
            portfolios will be calculated. For nested portfolios, the names must be
            given in the format "portfolio:held_portfolio", with multiple separators
            indicating deeper nesting. If nested portfolios are given, unpack must be
            True in the report settings.
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[PortfolioHoldingsReportAccessor]
            The task for the portfolio holdings report accessor.
        """


@docstrings_from_sync
class AsyncPortfolioHoldingsReportApi(
    AsyncReportApi[
        [list[str] | None, DateLike | None, DateLike | None],
        AsyncPortfolioHoldingsReportAccessor,
        PortfolioHoldingsReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncPortfolioHoldingsReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncPortfolioHoldingsReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        portfolio_names: list[str] | None,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncPortfolioHoldingsReportAccessor]: ...


class PortfolioHoldingsReportAccessorImpl(  # noqa: D101
    PortfolioHoldingsReportAccessor
):

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor


class AsyncPortfolioHoldingsReportAccessorImpl(  # noqa: D101
    AsyncPortfolioHoldingsReportAccessor
):

    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor



import abc
import datetime as dt
from typing import Any, Literal, cast

import polars as pl
from pydantic import Field

from bayesline.api._src._utils import docstrings_from_sync
from bayesline.api._src.equity.exposure_settings import ExposureSettings
from bayesline.api._src.equity.portfolioreport import (
    AsyncReportAccessorApi,
    ReportAccessorApi,
)
from bayesline.api._src.equity.report.accessor import (
    AsyncTypedReportAccessorApi,
    TypedReportAccessorApi,
)
from bayesline.api._src.equity.report.api import AsyncReportApi, ReportApi
from bayesline.api._src.equity.report.settings import ReportSettings
from bayesline.api._src.equity.universe_settings import UniverseSettings
from bayesline.api._src.tasks import AsyncTask, Task
from bayesline.api._src.types import DateLike, DNFFilterExpressions


class ExposureReportSettings(ReportSettings):
    """Settings for a exposure report."""

    report_type: Literal["exposure"] = Field(
        default="exposure", description="Report type discriminator"
    )
    universe_settings: UniverseSettings = Field()
    exposure_settings: ExposureSettings = Field()
    standardize_universe_settings: UniverseSettings | None = Field()


class ExposureReportAccessor(TypedReportAccessorApi):
    """Specific accessor for a exposure report."""

    @abc.abstractmethod
    def get_dates(self) -> list[dt.date]:
        """
        Get the dates for the report.

        Returns
        -------
        list[dt.date]
            The dates for the report.
        """

    @abc.abstractmethod
    def get(
        self, start_date: DateLike | None = None, end_date: DateLike | None = None
    ) -> pl.DataFrame:
        """
        Get the exposures for the report.

        Parameters
        ----------
        start_date: DateLike | None, default=None
            The start date of the report, inclusive.
        end_date: DateLike | None, default=None
            The end date of the report, inclusive.

        Returns
        -------
        pl.DataFrame
            The exposures for the report, with a date column and all factors as columns.
            The factor column names are of the form `factor_group^factor`.
        """


@docstrings_from_sync
class AsyncExposureReportAccessor(AsyncTypedReportAccessorApi):

    @abc.abstractmethod
    async def get_dates(self) -> list[dt.date]: ...  # noqa: D102

    @abc.abstractmethod
    async def get(  # noqa: D102
        self, start_date: DateLike | None = None, end_date: DateLike | None = None
    ) -> pl.DataFrame: ...


class ExposureReportApi(
    ReportApi[
        [DateLike | None, DateLike | None],
        ExposureReportAccessor,
        ExposureReportSettings,
    ]
):
    """API for a exposure report."""

    @abc.abstractmethod
    def get_maybe_cached(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> ExposureReportAccessor | None:
        """
        Get the cached exposure report accessor API if it is available.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        ExposureReportAccessor | None
            The cached exposure report accessor API, or None if the report is not cached.
        """

    @abc.abstractmethod
    def calculate(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> ExposureReportAccessor:
        """
        Calculate the exposure report.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        ExposureReportAccessor
            The exposure report accessor.
        """

    @abc.abstractmethod
    def calculate_as_task(
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> Task[ExposureReportAccessor]:
        """
        Calculate the exposure report as a task.

        Parameters
        ----------
        start_date: DateLike | None
            The start date of the report.
        end_date: DateLike | None
            The end date of the report.
        **kwargs: Any
            Additional keyword arguments.

        Returns
        -------
        Task[ExposureReportAccessor]
            The task for the exposure report accessor.
        """


@docstrings_from_sync
class AsyncExposureReportApi(
    AsyncReportApi[
        [DateLike | None, DateLike | None],
        AsyncExposureReportAccessor,
        ExposureReportSettings,
    ]
):

    @abc.abstractmethod
    async def get_maybe_cached(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncExposureReportAccessor | None: ...

    @abc.abstractmethod
    async def calculate(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncExposureReportAccessor: ...

    @abc.abstractmethod
    async def calculate_as_task(  # noqa: D102
        self,
        start_date: DateLike | None,
        end_date: DateLike | None,
        **kwargs: Any,
    ) -> AsyncTask[AsyncExposureReportAccessor]: ...


class ExposureReportAccessorImpl(ExposureReportAccessor):  # noqa: D101

    def __init__(self, accessor: ReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> ReportAccessorApi:  # noqa: D102
        return self._accessor

    def get_dates(self) -> list[dt.date]:  # noqa: D102
        return self.accessor.get_level_values(("date",)).to_series().dt.date().to_list()

    def get(  # noqa: D102
        self, start_date: DateLike | None = None, end_date: DateLike | None = None
    ) -> pl.DataFrame:
        filters: DNFFilterExpressions = []
        if start_date is not None:
            filters.append(("date", ">=", start_date))
        if end_date is not None:
            filters.append(("date", "<=", end_date))

        df = self.accessor.get_data(
            [],
            expand=("date", "asset_id"),
            value_cols=("exposure",),
            pivot_cols=("factor_group", "factor"),
            filters=filters or None,
        )
        return df.rename(lambda x: x.removesuffix("^exposure"))


class AsyncExposureReportAccessorImpl(AsyncExposureReportAccessor):  # noqa: D101

    def __init__(self, accessor: AsyncReportAccessorApi):
        self._accessor = accessor

    @property
    def accessor(self) -> AsyncReportAccessorApi:  # noqa: D102
        return self._accessor

    async def get_dates(self) -> list[dt.date]:  # noqa: D102
        df = await self.accessor.get_level_values(("date",))
        return cast(list[dt.date], df.to_series().dt.date().to_list())

    async def get(  # noqa: D102
        self, start_date: DateLike | None = None, end_date: DateLike | None = None
    ) -> pl.DataFrame:
        filters: DNFFilterExpressions = []
        if start_date is not None:
            filters.append(("date", ">=", start_date))
        if end_date is not None:
            filters.append(("date", "<=", end_date))

        df = await self.accessor.get_data(
            [],
            expand=("date", "asset_id"),
            value_cols=("exposure",),
            pivot_cols=("factor_group", "factor"),
            filters=filters or None,
        )
        return df.rename(lambda x: x.removesuffix("^exposure"))
